#!/usr/bin/env python
# -*- coding: utf-8 -*-

import logging
import json
import sys
from sys import argv
from libavg import avg, app, player as avgPlayer, geom
from libavg import logger
from math import fabs, radians
from World import World
from Tile import Tile
from Direction import Direction
from Level import *
from CheckpointManager import CheckpointManager
from hud.MovementControls import MovementControls
from hud.LightControls import LightControls
from hud.BalanceControls import BalanceControls
from hud.Minimap import Minimap
from event.BrokenSwampRails import BrokenSwampRails
from event.Portal import Portal
from event.FinalLootChest import FinalLootChest
from event.ItemHelpEvent import ItemHelpEvent
from event.BridgeEvent import BridgeEvent
from Map import Map
from item.Inventory import Inventory
from item.TurnBack import TurnBackRune
from item.Ping import PingRune
from menu.Menu import Menu
from menu.MenuTransition import MenuTransition
from StatisticsManager import StatisticsManager
from Player import Player
from menu.ScreenMessenger import ScreenMessenger
from menu.Button import Button
from menu.StyleDefinitions import StyleDefinitions

# Causes enumeration for a crash
CRASH_CAUSE_UNKNOWN = "unknown"
CRASH_CAUSE_BLOCKED_TILE = "blocked_tile"
CRASH_CAUSE_BALANCE = "balance"

# Setup Python Logger
hdlr = logging.StreamHandler()
# category is added as an extra formatting key by libavg
formatter = logging.Formatter('[%(asctime)s][%(levelname)s][%(category)s] : %(message)s')
hdlr.setFormatter(formatter)
pyLogger = logging.getLogger(__name__)
pyLogger.addHandler(hdlr)
pyLogger.propagate = False
pyLogger.level = logging.DEBUG

class Game(app.MainDiv):

    (resX,resY) = (None,None) # render resolution
    tileRes = avg.Point2D(117, 117) # size of one tile
    lightDistFactor = 3.0 # factor of how far the light can be pushed
    (borderX,borderY) = (5,4) # number of tiles to ignore while scrolling the camera
    borderDistances = None # Tuple containing absolute camera scrolling thresholds of screen coordinates
    (worldX,worldY) = (40,29) # size of world in tiles
    deskBorderSz = (300,300) # pixel size of the desk border
    dev = False # development mode status | set by "--dev" or "-d"
    disableEvents = False # No events will be triggered if set to True / set by "--no-events"
    noPlayerReset = False # If True, player will not be reset when hitting a blocked tile
    ACTION_KEYS = ('Up', 'Down')
    maxCheckpointLength = 15 # Maximum distance in tiles, the player will be resetted
    countdownNumber = 3 # Number, that is displayed by the after-event countdown
    pause = False # Bool used to block onFrame from applying the game logic, thus pausing it
    TELEPORT_ANIM_DURATION = 750 # ms
    MENU_FADEOUT_DURATION = 3500 # ms

    # Game Mode enum
    MODE_QUICKPLAY = 0
    MODE_CAMPAIGN = 1
    MODE_TUTORIAL = 2

    # Campaign level files
    __campaignDir = "../maps/campaign/"
    campaignLevels = [
        __campaignDir + "campaign_01.json"
        , __campaignDir + "campaign_02.json"
        , __campaignDir + "campaign_03.json"
        , __campaignDir + "campaign_04.json"
    ]

    # Tutorial level files
    __tutorialDir = "../maps/tutorial/"
    tutorialLevels = [
        __tutorialDir + "tutorial_01.json"
        , __tutorialDir + "tutorial_02.json"
        , __tutorialDir + "tutorial_03.json"
        , __tutorialDir + "tutorial_04.json"
        , __tutorialDir + "tutorial_05.json"
        , __tutorialDir + "tutorial_06.json"
    ]

    def onArgvParserCreated(self, parser):
        parser.add_option("-d", "--dev", dest="development", action="store_true", default=False, help="enable development features - advanced logging and windowed mode")
        parser.add_option("--draw-helpers", dest="drawHelpers", action="store_true", default=False, help="draw helpers nodes to visualize things for debugging")
        parser.add_option("--no-events", dest="disableEvents", action="store_true", default=False, help="ignore events completely")
        parser.add_option("--no-lighting", dest="noLighting", action="store_true", default=False, help="disable the black darkness overlay")
        parser.add_option("--no-reset", dest="noPlayerReset", action="store_true", default=False, help="do not reset the player when hitting a blocked tile. Useful for debugging.")
        parser.add_option("--level", dest="levelFile", default=None, help="specifies file with level state")
        parser.add_option("--no-menu", dest="noMenu", action="store_true", default=False, help="Do not start the menu at start time")

    def onArgvParsed(self, options, args, parser):
        print ""
        Game.dev = options.development
        Game.disableEvents = options.disableEvents
        Game.noPlayerReset = options.noPlayerReset
        avgPlayer.drawHelpers = options.drawHelpers
        Game.noLighting = options.noLighting
        Game.levelFile = options.levelFile
        Game.noMenu = options.noMenu
        if Game.dev:
            print "Development mode is enabled"
        if Game.disableEvents:
            print "Event detection is disabled"

    def onInit(self):
        # register logger
        avg.logger.removeStdLogSink()
        avg.logger.addSink(pyLogger)

        # enable INFO and DEBUG Severity for development mode
        if self.dev:
            logger.configureCategory(logger.Category.APP, logger.Severity.INFO)
            logger.configureCategory(logger.Category.APP, logger.Severity.DBG)

        avg.WordsNode.addFontDir("../assets/fonts")

        Tile.setTileSize(Game.tileRes.x, Game.tileRes.y)

        # Lower camera border size if needed
        borderSz = self.size - 2.0 * World.getPixelPos((Game.borderX,Game.borderY))
        if borderSz.x <= Game.tileRes.x or borderSz.y <= Game.tileRes.y:
            borderSz = avg.Point2D(max(borderSz.x, Game.tileRes.x), max(borderSz.y, Game.tileRes.y))
            Game.borderX = int(((self.size.x - borderSz.x) * 0.5) // Game.tileRes.x)
            Game.borderY = int(((self.size.y - borderSz.y) * 0.5) // Game.tileRes.y)
            logger.warning("Game border too big for tile size. Setting to maximum " + `(Game.borderX,Game.borderY)` + "...")

        # calculate Camera movement borders in pixel : right, left, down, up
        Game.borderDistances = (Game.resX - (Game.tileRes.x * (Game.borderX + 1)),
            Game.tileRes.x * (Game.borderX),
            Game.resY - (Game.tileRes.y * (Game.borderY + 1)),
            Game.tileRes.y * (Game.borderY))

        self.__avgPlayer = avg.Player.get()

        # Setup Main Div Nodes
        self.__desk = None

        # Init input system
        self.__keystates = dict.fromkeys(self.ACTION_KEYS, False)

        self.__avgPlayer.subscribe(avgPlayer.KEY_DOWN, self.__onKeyDown)
        self.__avgPlayer.subscribe(avgPlayer.KEY_UP, self.__onKeyUp)

        # How many levels (separate games) have been started yet
        self.__startedLevelCount = 0
        # By default, start new Game with random mazes
        self.__gameMode = Game.MODE_QUICKPLAY

        # Setup Statistics
        avgPlayer.statisticsManager = StatisticsManager()

        # Set default Balance Controls rotation behavior
        avgPlayer.balanceControlsDynamicRotation = False

        # prepare overlay
        self.__initOverlay()

        #determine, wether the event detection for the current tile should be skipped
        self.__currentTileEventHandled = False

        # If a new Level should be started on the next onFrame() Call
        self.__requestedNewLevel = False

        # Only show the menu when set in command line
        self.__menuNode = avg.DivNode(size=self.size, parent=self)
        self.__menu = Menu(self.__menuNode)
        self.__menuActive = True
        if Game.noMenu:
            self.__menu.hideMenu()
            self.__requestNewLevel()
        else:
            self.__menuTransition = MenuTransition(self)

    def __requestNewLevel(self):
        self.__requestedNewLevel = True

    # !!! DO NOT CALL THIS METHOD !!!
    # USE __requestNewLevel() to start a new level
    # This way, no strange method stack behavior appears
    def __startNewLevel(self):
        self.__requestedNewLevel = False
        self.__startedLevelCount += 1

        # clean up if needed
        if self.__desk is not None:
            self.__destroyLevel()

        # prepare GameMode specif level data / return if menu is to be showed
        if self.__setUpLevelData() != 0:
            return

        # Init base nodes
        self.__desk = avg.DivNode(parent=self)
        self.__deskBackground = avg.RectNode(filltexhref="../assets/frame.png", fillopacity=1.0, filltexcoord1=(0,0), strokewidth=0.0, parent=self.__desk)
        self.__board = avg.DivNode(pos=Game.deskBorderSz, parent=self.__desk)

        tileset = random.choice(['mine', 'swamp', 'underground', 'himalaya'])
        self.__level = Level(self.__board, (Game.worldX,Game.worldY), tileset, self.__levelData)
        
        # Resize the desk to match the world size
        worldSz = self.__level.getWorld().getSize()
        tileSz = Tile.getTileSize()
        self.__desk.size = avg.Point2D(worldSz[0] * tileSz[0] + 2 * Game.deskBorderSz[0], worldSz[1] * tileSz[1] + 2 * Game.deskBorderSz[1])
        self.__deskBackground.size = self.__desk.size
        self.__deskBackground.filltexcoord2 = avg.Point2D(worldSz[0],worldSz[1]) * avg.Point2D(Tile.getTileSize()) / avg.Point2D(200,200)
        self.__board.size = self.__desk.size - (2 * Game.deskBorderSz[0], 2 * Game.deskBorderSz[1])

        # Setup player
        self.__setupPlayer()

        # Init helpers
        if avgPlayer.drawHelpers:
            self.__borderHelper = avg.RectNode(pos=(Game.borderDistances[1], Game.borderDistances[3]), size=(Game.borderDistances[0] - Game.borderDistances[1] + Game.tileRes.x, Game.borderDistances[2] - Game.borderDistances[3] + Game.tileRes.y), strokewidth=2, color="FF0000", opacity=0.5, sensitive=False, parent=self)
            self.__checkpointManager = CheckpointManager(Game.maxCheckpointLength, self.__board)
        else:
            self.__checkpointManager = CheckpointManager(Game.maxCheckpointLength)

        self.__checkpointManager.addCheckpoint(self.__player.tilePos, self.__player.orientation, CheckpointManager.CHECKPOINT_START)

        # Setup Fullscreen event management
        self.__currentFullScreenEvent = None
        self.__currentFullScreenEventInited = False # was event.onInit() called?
        self.__eventNode = None

        # Setup inventory
        self.__inventory = Inventory(self, Game.resX - 300)
        self.__inventory.subscribe(Inventory.NEW_ITEM_COLLECTED, self.__onNewItemCollection)
        self.__inventory.subscribe(Inventory.ITEM_ACTIVATED, self.__onItemActivated)

        # Init controls
        self.__movementControls = MovementControls(self, self.__player, self.__inventory)
        self.__balanceControls = BalanceControls(self, self.__player)
        self.__lightControls = LightControls(self, self.__inventory)
        self.__minimap = Minimap(self.__level.getWorld(), self)

        height = 30
        width = 80
        self.__menuButton = Button(pos=(50, self.height - height), size=(width, height), text="Beenden", delay=1000, textColor=StyleDefinitions.FONT_COLOR, bgImageHref="../assets/menu/wood_5.png", parent=self)
        self.__menuButton.subscribe(Button.BUTTON_CLICKED_DELAY, self.__openMenu)

        self.__tutorialSkipButton = None

        # Initialize shading
        self.__initShading()

        self.__applyLevelSpecificSettings()

        # display Overlay on top
        self.reorderChild(self.__overlayDiv, self.getNumChildren()-1)
        self.__resetOverlay()

        if self.__gameMode == Game.MODE_CAMPAIGN:
            # Show fade-in animation into first level of the campaign
            if self.__startedLevelCount == 1:
                if self.__menuTransition:
                    self.__menuTransition.startFadeIn(None, 3500, 1500)

        if self.__gameMode == Game.MODE_TUTORIAL:
            self.__tutorialSkipButton = Button(pos=(50 + width + 10, self.height - height), size=(width, height), text="Überspringen", delay=500, textColor=StyleDefinitions.FONT_COLOR, bgImageHref="../assets/menu/wood_5.png", parent=self)
            self.__tutorialSkipButton.subscribe(Button.BUTTON_CLICKED_DELAY, self.__requestNewLevel)

        self.__requestedEvent = None

    def __setupPlayer(self):
        spawnPos = self.__level.getWorld().getSpawnPos()
        playerOrientation = self.__level.getWorld().getFreeDirs(spawnPos)[0]

        self.__player = Player(spawnPos, playerOrientation, "../assets/" + self.__level.getPlayerAsset(), self.__board)

        self.__playerPosBeforeEvent = self.__player.pos # Used to determine if the player moved during a fullscreen event
        self.__rotationAngle = 0 # Total angle to rotate to reach the target orientation
        self.__currentlyMoves = False
        self.__turnBackAtNextTile = False

    def __destroyLevel(self):
        # check if a fullscreen event is currently running
        if self.__currentFullScreenEvent is not None:
            self.__currentFullScreenEvent.disable()
            self.__currentFullScreenEvent = None
        if self.__eventNode is not None:
            self.__eventNode.unlink(True)
            self.__eventNode = None
        if self.__tutorialSkipButton is not None:
            self.__tutorialSkipButton.unlink(True)
            self.__tutorialSkipButton = None


        if self.__level:
            self.__level.destroy()

        self.__desk.unlink(True)
        self.__desk = None
        self.__menuButton.unlink(True)
        self.__menuButton = None
        self.__movementControls.destroy()
        self.__balanceControls.destroy()
        self.__lightControls.destroy()
        self.__minimap.destroy()
        self.__checkpointManager.destroy()
        self.__inventory.destroy()
        self.__player.unlink(True)
        self.__player = None

    def onFrame(self):
        if self.pause:
            return

        if self.__menuActive:
            self.__handleMenu()
            return

        if self.__requestedNewLevel:
            self.__startNewLevel()
            return

        if self.__requestedEvent:
            self.__onEventReached(self.__requestedEvent)
            self.__requestedEvent = None
            return

        if self.__isCampaign():
            if self.__mapPieceReached:
                if self.__mapPieces.hasShown():
                    self.__mapPieceReached = False
                    self.__requestNewLevel()
                return

        dt = self.__avgPlayer.getFrameDuration()
        isInEvent = (self.__currentFullScreenEvent is not None)
        isInFSEvent = (isInEvent and self.__currentFullScreenEvent.isFullscreenEvent())

        if isInFSEvent:
            # Add to FSE Time / has to be called before handleEvents to ensure event is not allready destroyed
            avgPlayer.statisticsManager.addValue(StatisticsManager.T_FSE, self.__currentFullScreenEvent.getEventName(), dt)

        if isInEvent:
            self.__handleFullScreenEvents()

        if not isInFSEvent:
            # Update world
            self.__level.onFrame(dt, self.__keystates)

            if not self.__overlayActive:
                # Update Inventory and item status
                self.__inventory.onFrame(dt)

                # Update light controls
                self.updateLightControls(dt)

                if not isInEvent:
                    # Update balance control
                    self.__balanceControls.onFrame(dt)
                    if abs(self.__player.balance) >= 1.0:
                        self.__onCrash(CRASH_CAUSE_BALANCE)

                    # Update movement control
                    self.__movementControls.onFrame(dt)
                    self.__player.onFrame(dt)

                    # Simulate player movement
                    self.__move()

                    if self.__isCampaign() and self.__mapPieceReached:
                        return

                self.__adjustCamera()

                #Move the Fog around if it exists
                if self.__level.getTileset() == 'swamp':
                    self.__moveFog()

                # Update minimap
                self.__minimap.onFrame(self.__player.pos)

                # add gameplayStatistics
                avgPlayer.statisticsManager.addValue(StatisticsManager.T_MC, str(self.__startedLevelCount), dt)

                # Draw Checkpoint helpers
                if avgPlayer.drawHelpers:
                    self.__checkpointManager.drawHelpers()

    def __showHUD(self, show=True):
        self.__lightControls.show(show)
        self.__movementControls.show(show)
        self.__balanceControls.show(show)
        self.__checkpointManager.show(show)
        self.__minimap.show(show)
        self.__inventory.show(show)
        if self.__menuButton is not None:
            self.__menuButton.active = show
        if self.__tutorialSkipButton is not None:
            self.__tutorialSkipButton.active = show

        if avgPlayer.drawHelpers:
            self.__borderHelper.active = show

    def __hideHUD(self):
        self.__showHUD(False)

    ##################################################################################
    #   Level Setup / Game Mode specific / Menu
    ##################################################################################

    def __displayMenu(self):
        self.__hideHUD()
        self.reorderChild(self.__menuNode, self.getNumChildren()-1) # display Menu above all other nodes
        self.__menu.showMenu()
        self.__menuActive = True

        # Stop existing fade animation, in case it is still running
        if self.__menuTransition:
            self.__menuTransition.abort()

    def __handleMenu(self):
        if self.__menuTransition and self.__menuTransition.isRunning():
            return

        self.__menu.onFrame()
        if self.__menu.startTutorial:
            self.__gameMode = Game.MODE_TUTORIAL
        elif self.__menu.startCampaign:
            self.__gameMode = Game.MODE_CAMPAIGN
            self.__mapPieceReached = False
            self.__mapPieces = Map(self)
            self.__mapPieces.hide()
        elif self.__menu.startQuickplay:
            self.__gameMode = Game.MODE_QUICKPLAY

        # Loop until there is interaction
        else:
            return

        # Fade out to black, then start the level
        if self.__menuTransition:
            self.__menuTransition.startFadeOut(self.__onMenuFadeOutFinished)

    def __onMenuFadeOutFinished(self):
        # Hide Menu and load level
        self.__menuActive = False
        self.__menu.reset()
        self.__menu.hideMenu()
        avgPlayer.statisticsManager.reset()
        self.__startedLevelCount = 0
        self.__requestNewLevel()

    def __setUpLevelData(self):
        self.__levelData = None
        if self.__gameMode == Game.MODE_CAMPAIGN:
            if len(Game.campaignLevels) >= self.__startedLevelCount:
                with open(Game.campaignLevels[self.__startedLevelCount - 1], 'r', 0) as fo:
                    self.__levelData = json.loads(fo.read())
            elif self.__startedLevelCount is len(Game.campaignLevels) + 1:
                # show final loot chest video
                self.__onEventReached(FinalLootChest())
                return 3
            else:
                self.__onCampaignFinished()
                return 1

        elif self.__gameMode == Game.MODE_TUTORIAL:
            if len(Game.tutorialLevels) >= self.__startedLevelCount:
                with open(Game.tutorialLevels[self.__startedLevelCount - 1], 'r', 0) as fo:
                    self.__levelData = json.loads(fo.read())
            else:
                self.__onTutorialFinished()
                return 2

        else: # asume Quickplay Mode as fallback
            if Game.levelFile:
                with open(Game.levelFile, 'r', 0) as fo:
                    self.__levelData = json.loads(fo.read())
        return 0

    def __applyLevelSpecificSettings(self):
        # disable visuals (and onFrame if HUD element is influencing gameplay passively) for unneeded HUDs in tutorials
        if self.__gameMode == Game.MODE_TUTORIAL:

            # activate per level
            if self.__startedLevelCount == 1:
                self.__minimap.active = False
                self.__darkness.active = False
                self.__lightControls.active = False
                self.__balanceControls.active = False
                self.__movementControls.active = False
            elif self.__startedLevelCount == 2:
                self.__minimap.active = False
                self.__darkness.active = False
                self.__lightControls.active = False
                self.__balanceControls.active = False
            elif self.__startedLevelCount == 3:
                self.__minimap.active = False
                self.__darkness.active = False
                self.__lightControls.active = False
            elif self.__startedLevelCount == 4:
                self.__minimap.active = False

            # only display active HUD elements
            self.__showHUD(False)
            self.__showHUD()


    def __onCampaignFinished(self):
        self.__displayMenu()
        self.__menu.displayCampaignFinishScreen()
        # TODO Implement winning screen / scoreboard ...

    def __onTutorialFinished(self):
        self.__displayMenu()

    def __isCampaign(self):
        return (True if self.__gameMode == Game.MODE_CAMPAIGN else False)



    ##################################################################################
    #   Rendering / Lighting / Shading
    ##################################################################################

    def __initShading(self):
        self.__fogLayers = []
        if self.__level.getTileset() == 'swamp':
            # Initialize Fog layers
            self.__fogLayers.append(avg.ImageNode(href="../assets/fog1.png", pos=(3840, 0), size=(Game.resX, Game.resY), sensitive=False, parent=self.__desk))
            self.__fogLayers.append(avg.ImageNode(href="../assets/fog1.png", pos=(1920,1200), size=(Game.resX, Game.resY), sensitive=False, parent=self.__desk))
            self.__fogLayers.append(avg.ImageNode(href="../assets/fog1.png", pos=(0,2400), size=(Game.resX, Game.resY), sensitive=False, parent=self.__desk))
        worldSz = (Game.worldX * Game.tileRes.x, Game.worldY * Game.tileRes.y)
        self.__darkness = avg.ImageNode(href="../assets/thedarkness.png", size=self.__desk.size, parent=self.__desk)
        self.__darkness.sensitive = False
        if self.__level.enableFlashlight():
            self.__darkness.maskhref = "../assets/light-nontransparent.png"
            self.__darkness.masksize = self.__lightControls.getLightSize()
            if Game.noLighting:
                self.__darkness.active = False



    ##################################################################################
    #   Player movement
    ##################################################################################

    # Returns True when the player can move on after calling this function or
    # False if a blocked tile or an event was encountered.
    def __move(self):
        # player reached the target tile
        if not self.__currentlyMoves:

            playerTilePos = self.__player.targetTilePos # same as self.__player.tilePos now

            # Make sure the player actually reached this tile.
            # Sometimes it may happen, that the player does not 100% reach its target tile
            # or is slightly off due to numeric issues.
            # Forcing him to the target position fixes possible crashes, but may result
            # in a skipping animation.
            if self.__player.tilePos != playerTilePos:
                print "Player did not reach target tile!"
            self.__player.pos = World.getPixelPos(playerTilePos)

            playerTilePosFrom = self.__player.originTilePos

            # detect free directions at current position
            freedirs = self.__level.getWorld().getFreeDirs(playerTilePos)

            # construct minimap
            for direction in freedirs:
                self.__minimap.addTile((direction.tuple()[0] + playerTilePos[0], direction.tuple()[1] + playerTilePos[1]))

            # reached Intersection?
            if len(freedirs) > 2:
                nextDir = self.__level.onCrossingReached(playerTilePos, playerTilePosFrom)
                if nextDir is None:
                    return False # we don't know where to go yet
                else:
                    move = nextDir.tuple()
                    self.__player.targetTilePos = (playerTilePos[0] + move[0], playerTilePos[1] + move[1])
                    self.__preparePlayerMovement(self.__player.tilePos, self.__player.targetTilePos)

                    self.__checkpointManager.addCheckpoint(self.__player.tilePos, self.__player.orientation, CheckpointManager.CHECKPOINT_INTERSECTION)
                    return True
            else:
                if not self.__currentTileEventHandled:
                    event = self.__level.onTileReached(playerTilePos, inventory=self.__inventory)
                    if event is not None and not Game.disableEvents:
                        self.__onEventReached(event)
                        return False
                else:
                    self.__currentTileEventHandled = False

                reachedTile = self.__level.getWorld().getTile(playerTilePos)

                # end level and restart if the final loot chest is reached
                if reachedTile.hasFeature("loot_chest"):
                    if self.__isCampaign(): # TODO Move out of this method, not movement related
                        logger.info("Campaign Level finished. Showing MapPieces.")
                        if not self.__mapPieceReached and self.__startedLevelCount < 5:
                            self.__handleMapPieces()
                        else:
                            self.__requestNewLevel() # Will load menu in start new level automatically
                        return False
                    else:
                        logger.info("Level finished. Starting new level.")
                        self.__requestNewLevel()
                        return False

                if reachedTile.hasFeature(Portal.FEATURE_NAME):
                    # Teleport the player to the other portal
                    portal = reachedTile.getFeature(Portal.FEATURE_NAME)
                    self.__teleportPlayerTo(portal.getOtherPortalPosition())
                    return False

                # Determine possible dirs we can move to
                possibleDirs = freedirs
                fromDir = self.__player.orientation
                if self.__turnBackAtNextTile:
                    # We just let the current orientation be the orientation we're coming from
                    self.__turnBackAtNextTile = False
                else:
                    fromDir = fromDir.inverted()

                if fromDir is not None and len(possibleDirs) is not 1 and fromDir in possibleDirs:
                    possibleDirs.remove(fromDir)

                move = possibleDirs[0].tuple()
                nextTilePos = (playerTilePos[0] + move[0], playerTilePos[1] + move[1])

                if self.__level.canMoveToTile(playerTilePos, nextTilePos):
                    self.__player.targetTilePos = nextTilePos

                    # initiate actual movement
                    self.__preparePlayerMovement(playerTilePos, self.__player.targetTilePos)
                    self.__checkpointManager.addCheckpoint(playerTilePos, self.__player.orientation, CheckpointManager.CHECKPOINT_REGULAR)
                    return True
                else:
                    self.__onCrash(CRASH_CAUSE_BLOCKED_TILE)
                    return False

        # Player is currently moving
        else:
            self.__movePlayerOnFrame()
            return True

    # oldPos, newPos - tile-positions as tuples
    def __preparePlayerMovement(self, oldPos, newPos):
        self.__currentlyMoves = True

        # set start angle correctly
        self.__player.angle = Player.getAngleFromOrientation(self.__player.lastOrientation)

        # percentual rotation will be added during __movePlayerOnFrame()
        self.__angleBeforeRotation = self.__player.angle

        if self.__player.orientation is not self.__player.lastOrientation:
            playerAngleDeg = math.degrees(self.__player.angle)

            endAngle = math.degrees(Player.getAngleFromOrientation(self.__player.orientation))
            tmpAngle = fabs(playerAngleDeg - endAngle) % 360

            # use shortest rotation direction
            if tmpAngle > 180:
                if playerAngleDeg > endAngle:
                    endAngle += 360
                else:
                    endAngle -= 360

            # how much the player needs to be rotated during the movement
            self.__rotationAngle = math.radians(endAngle - playerAngleDeg)

            # adjust player tilt according to relational rotation
            # positive -> right, negative-> left
            if self.__rotationAngle < 0:
                self.__player.tiltDir = 1
            else:
                self.__player.tiltDir = -1
        else:
            self.__rotationAngle = 0
            self.__player.tiltDir = 0

    # Simulate player forward
    #
    # The player is first simulated until reaching the next tile.
    # Then, the next tile is calculated and possible events fired (move())
    # and the player is simulated for the remaining time.
    def __movePlayerOnFrame(self):
        dt          = float(self.__avgPlayer.getFrameDuration()) * 0.001
        step        = dt * self.__player.speed
        remainVec   = self.__player.targetPos - self.__player.pos
        moveDir     = avg.Point2D(self.__player.targetTilePos) - avg.Point2D(self.__player.originTilePos)
        remainLn    = (remainVec.x * moveDir.x + remainVec.y * moveDir.y) # dot(remainVec, moveDir)

        # rotate player depending on movement progress on current tile
        distancePercentage = 1.0 - remainLn / Game.tileRes.x
        self.__player.tilt = self.__player.tiltDir * (1 - abs(0.5 - distancePercentage)) # larger acceleration when mid curve
        self.__player.angle = self.__angleBeforeRotation + distancePercentage * self.__rotationAngle

        if step >= remainLn:
            # Reset player to next tile
            orthoRemainVec = remainLn * moveDir
            self.__player.pos += orthoRemainVec
            self.__darkness.maskpos += orthoRemainVec

            # Update on reached tile
            self.__currentlyMoves = False
            if not self.__move():
                # We must not move further
                return

            # Simulate the rest time
            step -= remainLn

            # Recalculate movement vector
            moveDir = avg.Point2D(self.__player.targetTilePos) - avg.Point2D(self.__player.originTilePos)

        stepVec = step * moveDir
        self.__player.pos += stepVec

    # Moves the camera when the player gets close to the border of the screen
    def __adjustCamera(self):
        # camera follows the player and thus can only change its position on one axis
        # therefor, if one case is valid, the function can be left to reduce rendering time
        playerOffsetPos = self.__boardToScreenPos(self.__player.pos)

        shiftXR = playerOffsetPos.x - Game.borderDistances[0]
        if shiftXR > 0:
            self.__desk.x -= shiftXR
            return

        shiftXL = playerOffsetPos.x - Game.borderDistances[1]
        if shiftXL < 0:
            self.__desk.x -= shiftXL
            return

        shiftYD = playerOffsetPos.y - Game.borderDistances[2]
        if shiftYD > 0:
            self.__desk.y -= shiftYD
            return

        shiftYU = playerOffsetPos.y - Game.borderDistances[3]
        if shiftYU < 0:
            self.__desk.y -= shiftYU
            return

    # Move Fog around
    def __moveFog(self):
        for fog in self.__fogLayers:
            if fog.x <= self.__desk.width:
                fog.x += 3
            else:
                fog.x = 0


    # to - tuple tile position
    def __teleportPlayerTo(self, to):
        self.pause = True
        self.__balanceControls.showIndicator(False)
        playerImg = self.__player.imageNode
        self.__playerImgSizeBeforeTeleport = playerImg.size
        self.__playerImgPosBeforeTeleport = playerImg.pos
        self.__playerImgAngleBeforeTeleport = playerImg.angle
        fadeOutAnim = avg.ParallelAnim([
            avg.LinearAnim(playerImg, "size", Game.TELEPORT_ANIM_DURATION, playerImg.size, avg.Point2D(1,1)),
            avg.LinearAnim(playerImg, "pos", Game.TELEPORT_ANIM_DURATION, playerImg.pos, playerImg.pos + 0.5 * playerImg.size),
            avg.LinearAnim(playerImg, "angle", Game.TELEPORT_ANIM_DURATION, playerImg.angle, 18.0)
        ], None, lambda to=to: self.__onTeleportFadeOutFinished(to))
        fadeOutAnim.start()

    def __onTeleportFadeOutFinished(self, to):
        if self.__menuActive:
            return

        originTilePos = self.__player.tilePos

        self.__player.pos = World.getPixelPos(to)
        self.__player.speed = 0.0
        playerTilePos = self.__player.tilePos

        dirToOrigin = self.__level.determineShortestDirToCB(self.__player.tilePos, lambda tile, origin=originTilePos: tile.position == origin, 50)

        freeDirs = self.__level.getWorld().getFreeDirs(playerTilePos)
        if dirToOrigin in freeDirs:
            freeDirs.remove(dirToOrigin)

        nextDirTuple = (random.choice(freeDirs) if len(freeDirs) > 0 else dirToOrigin).tuple()
        self.__player.targetTilePos = (playerTilePos[0] + nextDirTuple[0], playerTilePos[1] + nextDirTuple[1])
        self.__player.resetLastOrientation()

        self.__preparePlayerMovement(playerTilePos, self.__player.targetTilePos)
        self.__adjustCamera()
        self.__updateLightMaskPosition()

        # Start fade-in animation
        playerImg = self.__player.imageNode
        fadeOutAnim = avg.ParallelAnim([
            avg.LinearAnim(playerImg, "size", Game.TELEPORT_ANIM_DURATION, playerImg.size, self.__playerImgSizeBeforeTeleport),
            avg.LinearAnim(playerImg, "pos", Game.TELEPORT_ANIM_DURATION, playerImg.pos, self.__playerImgPosBeforeTeleport),
            avg.LinearAnim(playerImg, "angle", Game.TELEPORT_ANIM_DURATION, playerImg.angle, self.__playerImgAngleBeforeTeleport)
        ], None, self.__onTeleportFadeInFinished)
        fadeOutAnim.start()

    def __onTeleportFadeInFinished(self):
        if self.__menuActive:
            return

        self.__balanceControls.showIndicator(True)
        self.pause = False


    ##################################################################################
    #   Crash Handling
    ##################################################################################

    def __onCrash(self, cause=CRASH_CAUSE_UNKNOWN):
        # Reset to the last cross
        if not Game.noPlayerReset:
            checkpoint = self.__checkpointManager.getLastValidCheckpoint(reset=True)
            checkpointPos = checkpoint[0]
            checkpointDir = checkpoint[1]

            self.updateLightControls(0)

            # Reset player position
            self.__player.pos = World.getPixelPos(checkpointPos)

            # Reset player target pos
            checkpointDirTuple = checkpointDir.tuple()
            self.__player.targetTilePos = (checkpointPos[0] + checkpointDirTuple[0], checkpointPos[1] + checkpointDirTuple[1])
            self.__player.angle = Player.getAngleFromOrientation(self.__player.orientation)
            self.__player.resetLastOrientation()
            self.__preparePlayerMovement(self.__player.tilePos, self.__player.targetTilePos)

            # Reset speed
            self.__movementControls.reset()

            # Reset balance
            self.__balanceControls.resetBalance()

            # Reset flashlight
            self.__darkness.maskpos = (self.__player.pos + Game.deskBorderSz + (self.__player.imageDiv.size - self.__darkness.masksize) * 0.5)

            # Show a crash overlay
            if cause is CRASH_CAUSE_BALANCE:
                self.__overlayText.text = "Entgleist"
            elif cause is CRASH_CAUSE_BLOCKED_TILE:
                self.__overlayText.text = "Gecrasht"
            else:
                self.__overlayText.text = "Gecrasht"

            avgPlayer.statisticsManager.addValue(StatisticsManager.C_CRASH , str(cause), 1)

            # Set crashed status and start the animation
            self.__hideHUD()

            self.__overlayActive = True
            self.__continueButton.opacity = 0.0
            self.__continueText.opacity = 0.0
            self.__overlayDiv.opacity = 1.0
            avg.Anim.fadeOut(self.__overlayDiv, 2000, self.__endCrash)
        else:
            if not cause is CRASH_CAUSE_BALANCE:
                self.__player.speed = MovementControls.SPEED_INIT

    def __endCrash(self):
            # Enable movement again
            self.__continueButton.opacity = 1.0
            self.__continueText.opacity = 1.0
            self.__showHUD()
            self.__resetOverlay()



    ##################################################################################
    #   Fullscreen events
    #
    # __onEventReached(): Shows name of event or directly calls __onDescriptionTimerFinished()
    #   __onDescriptionTimerFinished(): Shows description of the event or directly calls __startEvent()
    #     __startEvent(): Starts the actual event, onFrame() will then call __handleFullScreenEvents()
    # __handleFullScreenEvents(): Calls event.onFrame(). If it returns true, calls __onEventCompleted()
    #   __onEventCompleted(): Starts countdown and calls __onCountdownTimerFinished() for each tick
    ##################################################################################

    def __onEventReached(self, event):
        avgPlayer.statisticsManager.addValue(StatisticsManager.C_FSE, event.getEventName(), 1)
        self.__currentTileEventHandled = True
        self.__currentFullScreenEventInited = False
        if not self.__level.TILE_EVENTS[event.EVENT_NAME]['visitedOnce']:
            self.__overlayText.text = self.__level.TILE_EVENTS[event.EVENT_NAME]['shortDescription']
            self.__overlayActive = True
            self.__continueText.opacity = 0.0
            self.__continueButton.opacity = 0.0
            self.__overlayDiv.opacity = 1.0
            avg.Anim.fadeOut(self.__overlayText, 2000)
            self.__avgPlayer.setTimeout(2500, self.__onDescriptionTimerFinished)
            self.__currentFullScreenEvent = event
        else:
            self.__currentFullScreenEvent = event
            self.__onDescriptionTimerFinished()

        if self.__fogLayers and self.__currentFullScreenEvent is not None:
            if self.__currentFullScreenEvent.isFullscreenEvent():
                for layer in self.__fogLayers:
                    layer.opacity = 0.0

    def __onDescriptionTimerFinished(self):
        if not self.__currentFullScreenEvent:
            return

        if self.__currentFullScreenEvent.isFullscreenEvent():
            self.__eventNode = avg.DivNode(size=avg.Point2D(Game.resX, Game.resY), parent=self, sensitive=False)
            self.__level.getWorld().hide()
            if self.__player:
                self.__player.active = False
            self.__deskBackground.active = False
            self.__hideHUD()

        if self.__level.TILE_EVENTS[self.__currentFullScreenEvent.EVENT_NAME]['visitedOnce']:
            self.__startEvent()
        else:
            self.__descriptionMessenger = ScreenMessenger(self.__blackOverlay.pos, size=(self.resX, self.resY), headline=self.__level.TILE_EVENTS[self.__currentFullScreenEvent.EVENT_NAME]['shortDescription'], messageAlignment="center", message=self.__level.TILE_EVENTS[self.__currentFullScreenEvent.EVENT_NAME]['description'], parent=self, destroyOnFinish=True)
            self.__level.TILE_EVENTS[self.__currentFullScreenEvent.EVENT_NAME]['visitedOnce'] = True
            self.__descriptionMessenger.addSubscription(self.__startEvent)

        self.__overlayText.opacity = 1.0

    def __startEvent(self):
        self.__overlayDiv.sensitive = False
        self.__continueButton.opacity = 0.0
        self.__continueText.opacity = 0.0
        self.__overlayDiv.opacity = 0.0
        self.__overlayText.fontsize = Game.resX / 16
        self.__overlayText.pos = (Game.resX / 2, Game.resY / 2 - 100)
        self.__continueButton.sensitive = False
        if self.__eventNode:
            self.__eventNode.sensitive = True

        # Correctly handle events started before level creation (eg FinalLootChestEvent)
        if self.__player:
            self.__playerPosBeforeEvent = avg.Point2D(self.__player.pos)

        if self.__currentFullScreenEvent.getEventName() == BrokenSwampRails.EVENT_NAME:
            self.__currentFullScreenEvent.onInit(self.__eventNode, self.__player)
            self.__balanceControls.resetBalance()
            # Call onFrame() to update UI
            self.__balanceControls.onFrame(0)
        elif self.__currentFullScreenEvent.getEventName() == BrokenRail.EVENT_NAME:
            self.__currentFullScreenEvent.onInit(self.__eventNode, self.__player)
            if not self.__currentFullScreenEvent.onFrame(self.__avgPlayer.getFrameDuration(), self.__keystates):
                if self.__eventNode:
                    self.__eventNode.unlink(True)
                    self.__eventNode = None
                self.__level.getWorld().show()
                self.__deskBackground.active = True
                self.__currentFullScreenEvent = None

                self.__player.active = True
                self.__onCrash(CRASH_CAUSE_BALANCE)
        else:
            self.__currentFullScreenEvent.onInit(self.__eventNode)

        self.__currentFullScreenEventInited = True

    def __handleFullScreenEvents(self):
        if self.__currentFullScreenEvent == None or not self.__currentFullScreenEventInited:
            return

        # Call event.onFrame(), which returns True if the event was just completed
        if self.__currentFullScreenEvent.onFrame(self.__avgPlayer.getFrameDuration(), self.__keystates):
            self.__onEventCompleted()

    def __onEventCompleted(self):
        # Exit immediately if this is event was not invoked on the level
        if self.__currentFullScreenEvent.getEventName() == FinalLootChest.EVENT_NAME:
            self.__requestNewLevel()

        self.__currentFullScreenEvent.disable()
        self.__level.getWorld().show()
        self.__deskBackground.active = True

        self.__overlayActive = True
        self.__overlayDiv.opacity = 0.5
        self.__overlayText.text = "3"

        if self.__currentFullScreenEvent.requiresCountdownAfterCompletion():
            self.__avgPlayer.setTimeout(1000, self.__onCountdownTimerFinished)
            self.__avgPlayer.setTimeout(2000, self.__onCountdownTimerFinished)
            self.__avgPlayer.setTimeout(3000, self.__onCountdownTimerFinished)
        else:
            # Simulate finish of final countdown tick
            self.countdownNumber = 1
            self.__onCountdownTimerFinished()

        if self.__eventNode:
            self.__eventNode.unlink(True)
            self.__eventNode = None

        if self.__fogLayers:
            for layer in self.__fogLayers:
                layer.opacity = 1.0

        # Correctly handle events started before level creation (eg FinalLootChestEvent)
        if self.__player:
            self.__player.speed = 0.0
            self.__player.active = True

            addCheckpoint = False

            # Update playerPos and playerPosLast if necessary
            playerMovedDuringEvent = (self.__player.pos - self.__playerPosBeforeEvent).getNorm() > 0.001
            if playerMovedDuringEvent:
                # Determine the shortest dir to the event
                dirToEvent = self.__level.determineShortestDirToEvent(self.__player.tilePos, self.__currentFullScreenEvent.getEventName())

                # Position player to move away from event
                curTilePos = self.__player.tilePos
                freeDirs = self.__level.getWorld().getFreeDirs(curTilePos)
                if dirToEvent in freeDirs:
                    freeDirs.remove(dirToEvent)

                if len(freeDirs) > 0:
                    newOrientationTuple = freeDirs[0].tuple()
                else:
                    # We have no other choice to move the event again
                    newOrientationTuple = dirToEvent.tuple()

                self.__player.targetTilePos = (curTilePos[0] + newOrientationTuple[0], curTilePos[1] + newOrientationTuple[1])

                # Uncomment this to disable rotation animation after driving in the inverse direction
                #if dirToEvent.inverted().tuple() == newOrientationTuple:
                #    self.__player.resetLastOrientation()

                self.__player.speed = 0.0
                self.__preparePlayerMovement(curTilePos, self.__player.targetTilePos)
                addCheckpoint = True
            elif self.__currentFullScreenEvent.getEventName() == BridgeEvent.EVENT_NAME:
                addCheckpoint = True

            if addCheckpoint:
                self.__checkpointManager.addCheckpoint(self.__player.tilePos, self.__player.orientation, CheckpointManager.CHECKPOINT_INTERSECTION)

        self.__currentFullScreenEvent = None
        self.__showHUD()

    # Checks, if the countdown is finished or the number has to be changed
    def __onCountdownTimerFinished(self):
        if self.countdownNumber == 1:
            self.__resetOverlay()
        else:
            self.countdownNumber -= 1
        self.__overlayText.text = str(self.countdownNumber)

    # Start an Event on the next onFrame call
    def __requestEvent(self, event):
        self.__requestedEvent = event



    ##################################################################################
    #   Overlay
    ##################################################################################

    def __initOverlay(self):
        self.__overlayDiv = avg.DivNode(size=(Game.resX, Game.resY), opacity=0.0, sensitive=False, parent=self)
        self.__blackOverlay = avg.ImageNode(href="../assets/thedarkness.png", size=(Game.resX, Game.resY), opacity=1.0, sensitive=False, parent=self.__overlayDiv)
        self.__overlayText = avg.WordsNode(pos=(Game.resX / 2, Game.resY / 2 - 100), alignment="center", font=StyleDefinitions.FONT_HEADLINE, fontsize=Game.resX / 16, text="You Crashed", opacity=1.0, sensitive=False, parent=self.__overlayDiv)
        self.__continueText = avg.WordsNode(pos=self.__overlayText.pos + avg.Point2D(0, 300), alignment="center", font=StyleDefinitions.FONT_HEADLINE, fontsize=Game.resX / 24, text="weiter", opacity=1.0, parent=self.__overlayDiv)
        self.__continueButton = avg.RectNode(pos=self.__continueText.pos + avg.Point2D(-(Game.resX / 10), Game.resY / 200), size=(400, 100), opacity=1.0, color="FFFFFF", parent=self.__overlayDiv)
        self.__overlayActive = False # Used to block Player from moving

    def __resetOverlay(self):
        self.__overlayDiv.opacity = 0.0
        self.__overlayActive = False
        self.countdownNumber = 3


    def __handleMapPieces(self):
        self.__mapPieceReached = True
        self.__hideHUD()
        self.__desk.active = False
        self.__mapPieces.showPiece(self.__startedLevelCount)



    ##################################################################################
    #   Miscellaneous
    ##################################################################################

    def updateLightControls(self, dt):
        self.__lightControls.onFrame(dt)
        self.__updateLightMaskPosition()

    def __updateLightMaskPosition(self):
        lightDir = self.__lightControls.getDirection()
        self.__darkness.masksize = self.__lightControls.getLightSize()
        self.__darkness.maskpos = (self.__player.pos + Game.deskBorderSz + (self.__player.imageDiv.size - self.__darkness.masksize) * 0.5) + lightDir * self.lightDistFactor

    def __onNewItemCollection(self, args):
        self.__requestEvent(ItemHelpEvent(args))

    def __onItemActivated(self, item):
        if item.getItemName() == TurnBackRune.ITEM_NAME:
            self.__turnBackAtNextTile = True
        if item.getItemName() == PingRune.ITEM_NAME:
            self.__minimap.ping()

    def __openMenu(self):
        self.__destroyLevel()
        self.__displayMenu()

    def __onKeyDown(self, event):
        if event.keyname in self.ACTION_KEYS:
            self.__keystates[event.keyname] = True

        if not self.__menuActive:
            if event.keyname == 'F8':
                self.__saveLevelState()

            if event.keyname == 'P':
                self.pause = not self.pause

    def __onKeyUp(self, event):
        if event.keyname in self.ACTION_KEYS:
            self.__keystates[event.keyname] = False

        if not self.__menuActive:
            if event.keyname == 'L':
                self.__requestNewLevel()
            if event.keyname == '0':
                self.__openMenu()

    def __saveLevelState(self, filename='world.json'):
        jsonData = self.__level.serialize()
        with open(filename, 'w') as fo:
            fo.write(json.dumps(jsonData, separators=[',',':']))
            logger.info("Saved level state as '" + filename + "'")

    def __boardToScreenPos(self, p):
        return p + self.__board.pos + self.__desk.pos



if __name__ == '__main__':
    res = avgPlayer.getScreenResolution()
    Game.resX = res.x
    Game.resY = res.y
    windowX = res.x
    windowY = res.y
    fullScreen = "True"
    if "--dev" in argv or "-d" in argv:
        windowX *= 0.67
        windowY *= 0.67
        fullScreen = "False"

    app.App().run(Game(),app_resolution=str(Game.resX) + "x" + str(Game.resY),app_window_size=str(windowX) + "x" + str(windowY) ,app_fullscreen=fullScreen)
