#!/usr/bin/env python
# -*- coding: utf-8 -*-

from libavg import avg, player as avgPlayer, logger
from World import World
from Tile import Tile, TileFeature
from Direction import *
from Util import *
import math
from random import randint
from event.TileEvent import TileEvent
from event.PipesEvent import *
from event.LeverRiddleEvent import *
from event.MemoryEvent import MemoryEvent
from event.SpearTrap import *
from event.ObstacleTile import *
from event.BrokenSwampRails import *
from event.LockpickEvent import *
from event.TutorialEvent import *
from event.Avalanche import *
from event.Portal import *
from event.BridgeEvent import *
from event.FinalLootChest import FinalLootChest
from event.ItemHelpEvent import ItemHelpEvent
from event.Quicktime import QuicktimeEvent
from item.Petroleum import PetroleumTileFeature
from item.Brakefluid import BrakefluidTileFeature
from item.TurnBack import TurnBackRuneTileFeature
from item.Ping import PingRuneTileFeature
from event.BrokenRail import *
import random

class RailSwitch(TileFeature):
    FEATURE_NAME = "rail_switch"

    SWITCH_STATES = [
        { 'angle': 0, 'straight': False },
        { 'angle': 90, 'straight': False },
        { 'angle': 180, 'straight': False },
        { 'angle': 270, 'straight': False },
        { 'angle': 0, 'straight': True },
        { 'angle': 90, 'straight': True }
    ]

    def __init__(self, tile):
        super(RailSwitch, self).__init__(tile)
        self.__tilePos = tile.position
        self.__state = randint(0, len(RailSwitch.SWITCH_STATES) - 1)
        startAnchors = self.getAnchorPositions()
        self.setState(startAnchors[0], startAnchors[1])

        self.__dragContact = None
        self.__dragStartAnchor = None # RailSwitchAnchor

        # Create RailSwitchAnchors
        self.__anchors = []
        world = tile.getWorld()
        for freeDir in world.getFreeDirs(tile.position):
            freeDirTuple = freeDir.tuple()
            anchorPos = (tile.position[0] + freeDirTuple[0], tile.position[1] + freeDirTuple[1])
            anchorTile = world.getTile(anchorPos)
            if anchorTile and anchorPos in tile.getWorld().getMazePath():
                if not anchorTile.hasFeature(RailSwitchAnchor.FEATURE_NAME):
                    anchor = RailSwitchAnchor(anchorTile)
                    anchorTile.addFeature(anchor)
                else:
                    anchor = anchorTile.getFeature(RailSwitchAnchor.FEATURE_NAME)

                anchor.registerSwitch(self)
                self.__anchors.append(anchor)

        avgPlayer.getRootNode().subscribe(avg.Node.CURSOR_UP, self.__onCursorUp)

    def getFeatureName(self):
        return RailSwitch.FEATURE_NAME

    # Returns list of tuples of tile-positions of the anchor tiles
    #
    # Each switch has two anchors: The tiles on each input/output of the switch
    # Returns a list of the coordinates of the two anchor points A1 and A2
    #
    # +---+---+    +---+---+---+
    # | A2| \ |    | A2|-->|A1 |
    # +---+---+    +---+---+---+
    #     |A1 |
    #     +---+
    #
    # Parameters:
    #   state - If not a valid switch state index, the current state of this switch is used
    def getAnchorPositions(self, state=None):
        if state is None or state < 0 or state >= len(RailSwitch.SWITCH_STATES):
            state = self.__state

        stateInfo = RailSwitch.SWITCH_STATES[state]

        dirs = [(0,1),(0,-1)] if stateInfo['straight'] else [(1,0),(0,1)] # based on the curve asset
        for a in range(0, stateInfo['angle'], 90):
            dirs[:] = [rotateVec90CW(dir) for dir in dirs]

        return [(dirs[i][0] + self.__tilePos[0], dirs[i][1] + self.__tilePos[1]) for i in range(2)]

    # Sets the current state of the switch based on two anchors (tile-positions).
    # Will not do anything if the same anchor is passed twice (no match).
    #
    # Returns True if a state could be matched, False if not
    def setState(self, a1, a2):
        if a1 == a2:
            return False

        for state in range(len(RailSwitch.SWITCH_STATES)):
            anchors = self.getAnchorPositions(state)
            if anchors[0] == a1 and anchors[1] == a2 or anchors[0] == a2 and anchors[1] == a1:
                self.__state = state
                tile = self.getTile()
                stateInfo = RailSwitch.SWITCH_STATES[self.__state]
                tile.rotation = stateInfo['angle']
                tile.setTexture('path_straight' if stateInfo['straight'] else 'path_curve')

                return True

        return False

    def onStartDrag(self, anchor, event):
        if self.__dragContact is None:
            anchor.getTile().showHighlight()
            self.__dragContact = event.contact
            self.__dragStartAnchor = anchor

    def onAnchorHover(self, anchor, event):
        if self.__dragContact is not None and anchor != self.__dragStartAnchor:
            self.setState(anchor.getTile().position, self.__dragStartAnchor.getTile().position)
            for a in self.__anchors:
                a.getTile().showHighlight((a == anchor or a == self.__dragStartAnchor))

    def onDrop(self):
        self.__dragContact = None
        for a in self.__anchors:
            a.getTile().hideHighlight()

        self.__dragStartAnchor = None

    def __onCursorUp(self, event):
        if self.__dragContact and self.__dragContact == event.contact:
            self.onDrop()

    def onDestroy(self):
        for anchor in self.__anchors:
            anchorTile = anchor.getTile()
            if anchorTile:
                anchorTile.removeFeature(anchor.getFeatureName())
        self.__anchors = []

    def removeAnchor(self, anchor):
        if anchor in self.__anchors:
            self.__anchors.remove(anchor)


class RailSwitchAnchor(TileFeature):
    FEATURE_NAME = "rail_switch_anchor"

    def __init__(self, tile):
        super(RailSwitchAnchor, self).__init__(tile)
        self.__switches = []

    def getFeatureName(self):
        return RailSwitchAnchor.FEATURE_NAME

    # Adds another switch that can be changed by this anchor
    def registerSwitch(self, switch):
        tile = self.getTile()
        self.__switches.append({
            'switch': switch,
            'subscriptions': [
                tile.subscribe(avg.Node.CURSOR_DOWN, lambda e: switch.onStartDrag(self, e)),
                tile.subscribe(avg.Node.CURSOR_OVER, lambda e: switch.onAnchorHover(self, e)),
                tile.subscribe(avg.Node.CURSOR_UP, lambda e: switch.onDrop())
            ]})

    def onDestroy(self):
        tile = self.getTile()
        for switch in self.__switches:
            switch['switch'].removeAnchor(self)
            if tile:
                for subscription in switch['subscriptions']:
                    tile.unsubscribe(subscription)


class Level():
    TILE_EVENTS = {
        MemoryEvent.EVENT_NAME: { 'class': MemoryEvent, 'allowVertical': True, 'visitedOnce': False, 'shortDescription': "Magische Tür", 'description': "Eine magische Tür versperrt euch den Weg. <br/> Merkt euch die Reihenfolge <br/> der fünf aufleuchtenden Kreise, <br/> um die Tür zu öffnen." }
        ,PipesEvent.EVENT_NAME: { 'class': PipesEvent, 'allowVertical': False, 'visitedOnce': False, 'shortDescription': "Überschwemmung", 'description': "Ein Rohrbruch hat den Weg überschwemmt <br/> und hindert euch am weiterkommen. <br/> Stellt die Rohrteile so ein, <br/> dass das Wasser wieder hindurch fließen kann." }
        ,LeverRiddleEvent.EVENT_NAME: { 'class': LeverRiddleEvent, 'allowVertical': True, 'visitedOnce': False, 'shortDescription': "Hebelrätsel", 'description': "Bringt die Regler in die richtige Stellung,<br />um die Lampen leuchten zu lassen und<br />den dreistufigen Code zu knacken." }
        ,SpearTrap.FEATURE_NAME: { 'class': SpearTrap, 'allowVertical': True }
        ,ObstacleTile.FEATURE_NAME: { 'class': ObstacleTile, 'allowVertical': True }
        ,BrokenSwampRails.EVENT_NAME: { 'class': BrokenSwampRails, 'allowVertical': True, 'visitedOnce': True, 'tilesets': ['swamp'] }
        ,BridgeEvent.EVENT_NAME: { 'class': BridgeEvent, 'allowVertical': True, 'visitedOnce': False, 'shortDescription': "Brückenüberquerung", 'description': "Diese Brücke ist etwas beschädigt. <br/> Weicht den unsicheren Stellen aus, <br/> um sie zu überqueren." }
        ,PetroleumTileFeature.FEATURE_NAME: {'class': PetroleumTileFeature, 'allowVertical': True}
        ,BrakefluidTileFeature.FEATURE_NAME: {'class': BrakefluidTileFeature, 'allowVertical': True}
        ,TurnBackRuneTileFeature.FEATURE_NAME: { 'class': TurnBackRuneTileFeature, 'allowVertical': True }
        ,PingRuneTileFeature.FEATURE_NAME: {'class': PingRuneTileFeature, 'allowVertical': True}
        ,BrokenRail.EVENT_NAME: {'class': BrokenRail, 'allowVertical': True, 'visitedOnce': True}
        ,TutorialEvent.EVENT_NAME: {'class': TutorialEvent, 'visitedOnce': True, 'allowVertical': False, 'randomDeploy': False}
        ,LockpickEvent.EVENT_NAME: {'class': LockpickEvent, 'randomDeploy': True, 'allowVertical': True, 'visitedOnce': False, 'shortDescription': "Schloss Knacken", 'description': "Jeder Spieler kontrolliert eine Scheibe. <br/> Positioniert sie so, damit ihr in der Mitte <br/> auf das Schloss klicken könnt."}
        ,Avalanche.EVENT_NAME: {'class': Avalanche, 'allowVertical': True, 'visitedOnce': True, 'tilesets': ['himalaya'] }
        ,Portal.FEATURE_NAME: { 'class': Portal, 'allowVertical': True, 'visitedOnce': True }
        ,FinalLootChest.EVENT_NAME: {'class': FinalLootChest, 'visitedOnce': True, 'allowVertical': True, 'randomDeploy': False}
        ,ItemHelpEvent.EVENT_NAME: {'class': ItemHelpEvent, 'visitedOnce': True, 'allowVertical': True, 'randomDeploy': False}
        ,QuicktimeEvent.EVENT_NAME: {'class': QuicktimeEvent, 'visitedOnce': False, 'allowVertical': True, 'randomDeploy': True, 'shortDescription': 'Farbcode', 'description': 'Merkt euch die Reihenfolge der Farben und<br/>tippt sie so wieder an.<br/>Der Code hält 3 Versuche aus.' }
    }

    # Features that must not be overlayed by another event
    BLOCKED_FEATURES = [
        SpearTrapLever.FEATURE_NAME,
        WaterPuddle.FEATURE_NAME,
        ObstacleTile.FEATURE_NAME,
        BrokenSwampRailFeature.FEATURE_NAME,
        LootChest.FEATURE_NAME,
        Portal.FEATURE_NAME
    ]

    # Tileset is only used if no levelData was given
    def __init__(self, worldRootNode, (worldSzX,worldSzY), tileset, levelData=None):
        self.__worldDiv = worldRootNode
        self.__world = None
        self.__maze = []
        self.__lastReachedTile = None

        if levelData:
            self.__deserialize(levelData)
        else:
            self.__world = World(worldSzX, worldSzY, worldRootNode, tileset)
            self.__world.generateRandomMaze()
            self.__maze = self.__world.getMazePath()

        # Add switches
        for tilePos in self.__world.getMazePath():
            tile = self.__world.getTile(tilePos)

            # Add switch
            if self.__world.isCrossing(tilePos):
                tile.addFeature(RailSwitch(tile))

        if not levelData:
            self.__addRandomEvents()

    def destroy(self):
        self.__maze = []
        self.__world.destroy()
        self.__worldDiv.unlink(True)
        self.__worldDiv = None

    def getWorld(self):
        return self.__world

    def getPlayerAsset(self):
        return "player.png"

    def getTileset(self):
        return self.__world.getTileset()

    def enableFlashlight(self):
        return True

    # Returns the level serialized as a dictionary
    def serialize(self):
        worldData = self.__world.serialize()
        worldData['events'] = {}
        for pos in self.__world.getMazePath():
            tile = self.__world.getTile(pos)
            for eventName in Level.TILE_EVENTS:
                eventFeature = tile.getFeature(eventName)
                if not eventFeature:
                    eventFeature = tile.getFeature(TileEvent.FEATURE_NAME)
                    if not eventFeature or eventFeature.getEventName() != eventName:
                        continue

                # If this is an event that was spread over neigbor tiles,
                # it's tile is not the same as the current tile
                if eventFeature.getTile().position != pos:
                    continue

                eventData = serializeVector(pos)
                if eventName == SpearTrap.FEATURE_NAME:
                    leverPos = eventFeature.getLever().getTile().position
                    eventData['levers'] = [serializeVector(leverPos)]

                if not eventName in worldData['events']:
                    worldData['events'][eventName] = []
                worldData['events'][eventName].append(eventData)

        return {
            'world': worldData,
            'tileset': self.getTileset(),
            'spawn': serializeVector(self.__world.getSpawnPos()),
            'objective': serializeVector(self.__world.getObjectivePos())
        }

    # Reconstructs the level state from the given level data dictionary
    def __deserialize(self, levelData):
        # Will read the tileset from data, or default to mine of none was found
        if 'tileset' in levelData:
            tileset = levelData['tileset']
        else:
            tileset = 'mine'

        if 'world' in levelData:
            worldData = levelData['world']
            if not 'size' in worldData:
                logger.error("Cannot load world: Missing world.size")
                return

            worldSz = deserializeVector(worldData['size'])
            if worldSz is None:
                logger.error("Cannot load world: Invalid world.size");
                return

            self.__world = World(worldSz[0], worldSz[1], self.__worldDiv, tileset)
            self.__world.deserialize(levelData['world'])
            self.__maze = self.__world.getMazePath()

            if 'events' in worldData:
                eventsData = worldData['events']
                for eventName in Level.TILE_EVENTS:
                    self.__deserializeEventsOfType(eventsData, eventName)
        else:
            logger.error("World data missing in loaded level data")
            return

        if 'spawn' in levelData:
            spawnPos = deserializeVector(levelData['spawn'])
            if spawnPos is not None:
                self.__world.setSpawnPos(spawnPos)
            else:
                logger.warning("Invalid spawn data in level")

        if 'objective' in levelData:
            objPos = deserializeVector(levelData['objective'])
            if objPos is not None:
                self.__world.setObjectivePos(objPos)
            else:
                logger.warning("Invalid objective data in level")

    def __deserializeEventsOfType(self, eventsData, type):
        if not type in eventsData:
            return

        event = Level.TILE_EVENTS[type]
        if not event:
            return

        for eventData in eventsData[type]:
            pos = deserializeVector(eventData)
            if pos is not None:
                tile = self.__world.getTile(pos)
                if not tile.hasFeature(TileEvent.FEATURE_NAME):
                    # Load lever pos for SpearTrap Feature
                    if type == SpearTrap.FEATURE_NAME and 'levers' in eventData:
                        leversData = eventData['levers']

                        # Use first lever in list only
                        if len(leversData) > 0:
                            leverPos = deserializeVector(leversData[0])
                            if leverPos is not None:
                                tile.addFeature(event['class'](tile, leverPos))

                    elif type == Portal.FEATURE_NAME and 'other' in eventData:
                        otherPos = deserializeVector(eventData['other'])
                        tile.addFeature(event['class'](tile, otherPos))

                    elif type == TutorialEvent.EVENT_NAME:
                        # Special treatment for Tutorial Tilefeatures which need an additional id
                        tutID = 0
                        if not 'tutID' in eventData:
                            logger.error("No ID for TutorialEvent @" + str(pos) + " found, fallback to ID 0!")
                        else:
                            tutID = eventData['tutID']
                        tile.addFeature(event['class'](tile, tutID))

                    else:
                        tile.addFeature(event['class'](tile))

    # (x,y) - Tuple of tile-coordinates of the crossing tile
    # Returns the Direction in which way to go next or None if the Level still needs time to decide which way to go
    def onCrossingReached(self, (x,y), cameFrom):
        tile = self.__world.getTile((x,y))
        switch = tile.getFeature(RailSwitch.FEATURE_NAME)
        if not switch:
            return self.__world.getFreeDirs((x,y))[0]

        anchors = switch.getAnchorPositions()

        # Stop if switch is in incorrect position
        if not cameFrom in anchors:
            return None

        anchors.remove(cameFrom)
        next = anchors[0]
        dir = Direction.fromTuple((next[0] - x, next[1] - y))
        if dir in self.__world.getFreeDirs((x,y)):
            return dir
        else:
            return None

    # (x,y) - Tuple of tile-coordinates of the reached tile
    # Returns None or a fullscreen event instance, if the tile is an fullscreen-event-triggering tile
    def onTileReached(self, (x,y), **kwargs):
        tile = self.__world.getTile((x,y))
        if tile is not self.__lastReachedTile:
            if self.__lastReachedTile:
                self.__lastReachedTile.onPlayerLeft()
            self.__lastReachedTile = tile
            self.__lastReachedTile.onPlayerEntered(**kwargs)

        event = tile.getFeature(TileEvent.FEATURE_NAME)
        if event and not event.isDisabled():
            return event
        else:
            return None

    def canMoveToTile(self, (fromX,fromY), (toX,toY)):
        toTile = self.__world.getTile((toX, toY))
        if toTile.isBlocked():
            return False

        spearTrap = toTile.getFeature(SpearTrap.FEATURE_NAME)
        if spearTrap and spearTrap.isActivated():
            return False

        obstacleTile = toTile.getFeature(ObstacleTile.FEATURE_NAME)
        if obstacleTile and obstacleTile.isActivated():
            return False

        switch = toTile.getFeature(RailSwitch.FEATURE_NAME)
        if switch is None:
            return True

        anchors = switch.getAnchorPositions()
        if not (fromX,fromY) in anchors:
            return False

        anchors.remove((fromX,fromY))
        next = anchors[0]
        dir = Direction.fromTuple((next[0] - toX, next[1] - toY))
        return (dir in self.__world.getFreeDirs((toX,toY)))

    def onFrame(self, dt, keystates):

        pass

    def __addRandomEvents(self):
        for tilePos in self.__world.getMazePath():
            tile = self.__world.getTile(tilePos)
            vertical = tile.getRotation() in [0, 180]
            if not tile.isEvent():
                continue
            if self.__world.isCrossing(tilePos) or self.__world.isNextToCrossing(tilePos):
                continue
            if tile.hasAnyFeature(Level.BLOCKED_FEATURES):
                continue

            # Determine events that are suitable
            possibleEvents = []
            for eventName in Level.TILE_EVENTS.keys():
                event = Level.TILE_EVENTS[eventName]
                # To constraint the creation further, add an elif branch here
                if 'tilesets' in event and not self.getTileset() in event['tilesets']:
                    continue
                elif vertical and not event['allowVertical']:
                    continue
                elif 'randomDeploy' in event and not event['randomDeploy']:
                    continue
                else:
                    possibleEvents.append(eventName)

            eventName = random.choice(possibleEvents)
            event = Level.TILE_EVENTS[eventName]

            tile.addFeature(event['class'](tile))

    # Determines the Direction in which the player can move next to
    # avoid the moving to a specific event with given name again.
    def determineShortestDirToEvent(self, startPos, eventName):
        def critfn(tile):
            eventFeature = tile.getFeature(TileEvent.FEATURE_NAME)
            return (eventFeature is not None and eventFeature.getEventName() == eventName)

        return self.determineShortestDirToCB(startPos, critfn)

    # Determines the Direction in which the player can move next to
    # avoid moving to a tile where the callback function returns true.
    #
    # bool fn(Tile tile) - the callback function
    def determineShortestDirToCB(self, startPos, fn, maxdist=15):
        world = self.__world
        shortestDir = None
        shortestDirDist = sys.maxint
        shortestDirFound = False # actually found in this direction
        for startDir in world.getFreeDirs(startPos):
            # In case we will have no result, use the first free dir
            if shortestDir is None:
                shortestDir = startDir

            startDirTuple = startDir.tuple()
            startDirPos = (startPos[0] + startDirTuple[0], startPos[1] + startDirTuple[1])
            open, closed = [startDirPos], [startPos]
            found = False
            while len(open) > 0 and len(closed) < maxdist:
                curPos = open.pop()
                closed.append(curPos)
                tile = world.getTile(curPos)
                if fn(tile):
                    found = True
                    break

                freeDirs = world.getFreeDirs(curPos)
                for freeDir in freeDirs:
                    freeDirTuple = freeDir.tuple()
                    freePos = (curPos[0] + freeDirTuple[0], curPos[1] + freeDirTuple[1])
                    if not freePos in closed:
                        open.append(freePos)

            dist = len(closed)
            closer = dist <= shortestDirDist
            if (not shortestDirFound and (found or closer)) or (shortestDirFound and found and closer):
                shortestDir = startDir
                shortestDirDist = dist
                shortestDirFound = found

        return shortestDir
