#!/usr/bin/env python
# -*- coding: utf-8 -*-

from libavg import avg, logger, player as avgPlayer
from abc import ABCMeta, abstractmethod
import math

class TileFeature:
    __metaclass__ = ABCMeta

    def __init__(self, tile):
        self.__tile = tile

    def getTile(self):
        return self.__tile

    def destroy(self):
        self.onDestroy()
        self.__tile = None

    # Can be overriden to handle destruction before self.__tile will be reset
    def onDestroy(self):
        pass

    @abstractmethod
    def getFeatureName(self):
        return "unknown_feature"

    # Called once everytime the player enters the tile of that TileFeature
    def onPlayerEnteredTile(self, **kwargs):
        pass

    # Called once when the player leaves the tile of that TileFeature
    def onPlayerLeftTile(self):
        pass


class Tile(avg.ImageNode):

    textures = {
        'ground': 'ground.png',
        'path_curve': {
            0: 'path_curve_0.png',
            90: 'path_curve_90.png',
            180: 'path_curve_180.png',
            270: 'path_curve_270.png',
        },
        'path_straight': {
            0: 'path_straight_0.png',
            90: 'path_straight_90.png',
            180: 'path_straight_0.png',
            270: 'path_straight_90.png',
        },
        'obstacle_large': 'obstacle_large.png',
        'obstacle_medium': 'obstacle_medium.png',
        'obstacle_small': 'obstacle_small.png',
        'broken_rail': {
            0: 'path_straight_0_broken.png',
            90: 'path_straight_90_broken.png',
            180: 'path_straight_180_broken.png',
            270: 'path_straight_270_broken.png',
        },
        'broken_swamp_rail': 'broken_tile_0.png',
        'dead_end': {
            0: 'dead_end_0.png',
            90: 'dead_end_90.png',
            180: 'dead_end_180.png',
            270: 'dead_end_270.png'
        },
        'chest_loot' : '../chest_loot.png'
    }

    # texture - an asset id from the Tile.textures dict
    @staticmethod
    def getTextureAssetHref(tileset, texture, angle=0):
        asset = Tile.textures[texture]
        if isinstance(asset, dict) and (angle in asset):
            asset = asset[angle]

        return "../assets/" + tileset + "/" + asset


    __tileSize = (160,160)

    # textureType - an asset id from the Tile.textures dict
    # tileset - the name of the tileset to use (assets subdirectory)
    # pos - position of the tile in Pixels
    # parent - parent node for this Tile
    # isFullscreenEvent - true if this Tile triggers a FullscreenEvent
    def __init__(self, textureType, tileset, pos, parent, isFullscreenEvent, world, **kwargs):
        self.__world = world
        self.__features = []
        self.__freeDirections = []
        self.__texture = textureType
        self.__tileset = tileset
        self.__isBlocked = False
        self.__isEvent = isFullscreenEvent
        self.__angle = 0 # degrees
        self.__highlightImg = None # if None, the tile is not highlit

        super(Tile, self).__init__(pos=pos, size=self.__tileSize, **kwargs)
        self.registerInstance(self, parent)
        self.setTexture(textureType)

        if avgPlayer.drawHelpers:
            self.elementoutlinecolor = "553355"

    def unlink(self, gc):
        while len(self.__features) > 0:
            self.__features.pop().destroy()

        super(Tile, self).unlink(gc)

    @staticmethod
    def getTileSize():
        return Tile.__tileSize

    @staticmethod
    def setTileSize(x, y):
        Tile.__tileSize = (x,y)


    # Returns the World, this tile was created in
    def getWorld(self):
        return self.__world


    # Adds the given TileFeature instance to this tile.
    # Returns:
    #   False if the given feature is not a TileFeature or the tile already has this feature.
    #   True otherwise
    def addFeature(self, feature):
        if not isinstance(feature, TileFeature):
            logger.error("Cannot add feature to tile @" + `self.position` + ": not a TileFeature!")
            return False

        if self.hasFeature(feature.getFeatureName()):
            logger.debug("Tile.addFeature('" + `feature.getFeatureName()` + "'): Tile @" + `self.position` + " already has such a feature")
            return False

        self.__features.append(feature)
        return True

    # Removes the TileFeature with the given feature name from this tile
    # Returns:
    #   False if this tile does not have such a feature
    #   True if the feature was found and removed
    def removeFeature(self, featureName, destroy=True):
        feature = self.getFeature(featureName)
        if feature is not None:
            if destroy:
                feature.destroy()
            self.__features.remove(feature)
            return True
        else:
            return False

    # Removes and destroys all features of this tile
    def removeAllFeatures(self):
        for feature in self.__features:
            feature.destroy()
        self.__features = []

    # Returns the TileFeature instance with the given feature name or None if this tile
    # does not have such a feature
    def getFeature(self, featureName):
        for feature in self.__features:
            if feature.getFeatureName().lower() == featureName.lower():
                return feature
        return None

    # Returns True if this tile has a feature with the given feature name (case insensitive)
    def hasFeature(self, featureName):
        return self.getFeature(featureName) is not None

    # Returns True if this tile has any of the features given by their name in the list (case insensitive)
    # If featureNameList is None or not given, True is returned if the tile has any feature.
    def hasAnyFeature(self, featureNameList=None):
        if featureNameList is None:
            return len(self.__features) > 0

        for featureName in featureNameList:
            if self.hasFeature(featureName):
                return True
        return False


    def getTexture(self):
        return self.__texture

    def setTexture(self, textureType):
        if not textureType:
            self.href = ""
            self.active = False
            return

        if not textureType in Tile.textures:
            logger.error("Cannot set texture to tile @" + `self.position` + ": '" + textureType + "' is not a valid texture type!")
            return

        self.__texture = textureType
        self.href = Tile.getTextureAssetHref(self.__tileset, self.__texture, self.__angle)
        self.active = True

    texture = property(getTexture, setTexture)


    def getTileset(self):
        return self.__tileset

    def setTileset(self, tileset):
        self.__tileset = tileset
        if self.__texture:
            self.href = Tile.getTextureAssetHref(self.__tileset, self.__texture, self.__angle)

    tileset = property(getTileset, setTileset)


    def isEvent(self):
        return self.__isEvent

    def setEvent(self, isevent=True):
        self.__isEvent = isevent

    isevent = property(isEvent, setEvent)


    def getPixelPos(self):
        return (self.x, self.y)

    def setPixelPos(self, (x,y)):
        self.x = x
        self.y = y


    def getPos(self):
        return (int(self.x // Tile.getTileSize()[0]), int(self.y // Tile.getTileSize()[1]))

    def setPos(self, (x,y)):
        self.x = x * Tile.getTileSize[0]
        self.y = y * Tile.getTileSize[1]

    position = property(getPos,setPos)


    def getRotation(self):
        return self.__angle

    # angle - (degrees) should be a multiple of 90. If not, will use the lower multiple of 90
    def setRotation(self, angle):
        self.__angle = (angle - angle % 90)
        self.setTexture(self.__texture)

    rotation = property(getRotation, setRotation)


    def block(self):
        self.__isBlocked = True

    def unblock(self):
        self.__isBlocked = False

    def isBlocked(self):
        return self.__isBlocked


    def showHighlight(self, show=True):
        if show:
            if not self.__highlightImg:
                self.__highlightImg = avg.ImageNode(
                    pos=self.pos,
                    size=self.size,
                    href='../assets/switch-anchor-highlight.png',
                    angle=math.radians(self.rotation),
                    sensitive=False,
                    parent=self.parent)
            self.__highlightImg.active = True
        else:
            if self.__highlightImg:
                self.__highlightImg.active = False

    def hideHighlight(self):
        self.showHighlight(False)


    def onPlayerEntered(self, **kwargs):
        for feature in self.__features:
            feature.onPlayerEnteredTile(**kwargs)

    def onPlayerLeft(self):
        for feature in self.__features:
            feature.onPlayerLeftTile()




class LootChest(TileFeature):
    FEATURE_NAME = "loot_chest"

    def __init__(self, tile):
        super(LootChest, self).__init__(tile)

        self.__chestImg = avg.ImageNode(
            href=Tile.getTextureAssetHref(tile.tileset, 'chest_loot', tile.rotation),
            pos=tile.pos,
            size=tile.size,
            angle=math.radians(tile.rotation),
            parent=tile.parent)

    def onDestroy(self):
        self.__chestImg.unlink(True)
        self.__chestImg = None

    def getFeatureName(self):
        return LootChest.FEATURE_NAME
