#!/usr/bin/env python
# -*- coding: utf-8 -*-

from libavg import avg, app, logger
from random import *
import sys
from Tile import Tile, LootChest
from item.Petroleum import PetroleumTileFeature
from item.Brakefluid import BrakefluidTileFeature
from Direction import *
from Util import *
import math

class World():

    __gridSize = 5
    DEFAULT_BACKGROUND_TEXTURE_SIZE = (500,500)

    # width, height - in tiles
    def __init__(self, width, height, parentNode, tileset, **kwargs):
        self.__parentNode = parentNode
        self.__tileset = tileset
        self.__width = width
        self.__height = height
        self.__minEventDistance = 10 # minimal Distance between events

        self.__background = avg.RectNode(
            pos=(0,0),
            size=World.getPixelPos((width,height)),
            filltexhref=Tile.getTextureAssetHref(tileset, "ground"),
            fillopacity=1.0,
            filltexcoord1=(0,0),
            strokewidth=0.0,
            parent=self.__parentNode)

        self.__setBackgroundTextureSize(World.DEFAULT_BACKGROUND_TEXTURE_SIZE)

        self.__tiles = {(x,y): Tile('', tileset, World.getPixelPos((x,y)), self.__parentNode, False, self) for x in range(self.__width) for y in range(self.__height)}
        self.__path = [] # list of tile-positions of the path tiles
        self.__freeDirectionCache = {}
        self.__spawnPos = (1,1)
        self.__objectivePos = (2,1)

    def destroy(self):
        while len(self.__tiles) > 0:
            self.__tiles.popitem()[1].unlink(True)
        self.__tiles.clear()
        self.__path = []
        self.__background.unlink(True)
        self.__background = None

    # Generates a random maze world
    def generateRandomMaze(self):
        self.__setPath(self.__calculateRandomPath())
        self.setSpawnPos(self.__path[0])
        self.setObjectivePos(self.__path[-1])

    # Reconstructs the given serialized world state
    # WARNING: This method assumes that the world size is big enough to hold all path tiles!
    # worldData - a dictionary output by World.serialize()
    def deserialize(self, worldData):
        # Path itself
        if not 'path' in worldData:
            logger.error("Cannot deserialize world: Missing path data")
            return

        path = []
        numInvalid = 0
        for tileInfo in worldData['path']:
            if 'x' in tileInfo and 'y' in tileInfo:
                path.append((tileInfo['x'], tileInfo['y']))
            else:
                numInvalid += 1

        if numInvalid > 0:
            logger.error(`numInvalid` + " corrupt tiles in world data")

        self.__setPath(path, False)

        # Background texture size
        bgTexSz = World.DEFAULT_BACKGROUND_TEXTURE_SIZE
        if 'backgroundTextureSize' in worldData:
            bgTexSzData = worldData['backgroundTextureSize']
            if 'x' in bgTexSzData and 'y' in bgTexSzData:
                bgTexSz = ((bgTexSzData['x'],bgTexSzData['y']))

        self.__setBackgroundTextureSize(bgTexSz)


    # Returns the world state serialized as a dict that can be loaded by deserialize() later
    def serialize(self):
        # Note: Events are not added here, as the world itself does
        #       not now anything about events or features.
        return {
            'size': serializeVector((self.__width,self.__height)),
            'backgroundTextureSize': serializeVector(self.getBackgroundTextureSize()),
            'path': [serializeVector(pos) for pos in self.__path]
        }


    def hide(self):
        self.__parentNode.active = False

    def show(self):
        self.__parentNode.active = True

    # Returns the tile-position of the starting tile in the maze as a tuple
    def getSpawnPos(self):
        return self.__spawnPos

    def setSpawnPos(self, (x,y)):
        self.__spawnPos = (x,y)

    # Returns the tile-position of the last tile in the maze
    def getObjectivePos(self):
        return self.__objectivePos

    def setObjectivePos(self, (x,y)):
        prevObjectiveTile = self.getTile(self.__objectivePos)
        if prevObjectiveTile:
            prevObjectiveTile.removeFeature(LootChest.FEATURE_NAME)

        objectiveTile = self.getTile((x,y))
        if objectiveTile:
            self.__objectivePos = (x,y)
            objectiveTile.addFeature(LootChest(objectiveTile))

    # Returns list of tile-positions in the order of the maze path
    def getMazePath(self):
        return self.__path

    # Returns a Tuple of the worlds width and height in tiles
    def getSize(self):
        return (self.__width, self.__height)

    # (w,h) - screen-space size in pixels
    def __setBackgroundTextureSize(self, (w,h)):
        self.__background.filltexcoord2 = avg.Point2D(self.__width,self.__height) * avg.Point2D(Tile.getTileSize()) / avg.Point2D(w,h)

    # Returns tuple (w,h) describing the size of one background texture tile size
    def getBackgroundTextureSize(self):
        bgTexSz = (avg.Point2D(self.__width,self.__height) * avg.Point2D(Tile.getTileSize())) / self.__background.filltexcoord2
        return (bgTexSz.x,bgTexSz.y)

    # Converts tile-position to pixel position
    # Returns: Tuple of pixel coordinates of the given tile
    @staticmethod
    def getPixelPos((x,y)):
        return avg.Point2D(x * Tile.getTileSize()[0], y * Tile.getTileSize()[1])

    # Converts pixel position to tile position by rounding off
    # Returns: Tuple of tile-coordinates of the tile at given pixel position
    @staticmethod
    def getTilePos((x,y)):
        return (int(x // Tile.getTileSize()[0]), int(y // Tile.getTileSize()[1]))

    # (x,y) - Tile-Position of tile to check directions for
    # Returns: List of Direction's in which the player may move at the given tile
    def getFreeDirs(self, (x,y)):
        if (x,y) in self.__freeDirectionCache:
            return self.__freeDirectionCache[(x,y)][:]

        dirs = []
        if (x,y+1) in self.__path: dirs.append(Direction.DOWN)
        if (x+1,y) in self.__path: dirs.append(Direction.RIGHT)
        if (x,y-1) in self.__path: dirs.append(Direction.UP)
        if (x-1,y) in self.__path: dirs.append(Direction.LEFT)

        self.__freeDirectionCache[(x,y)] = dirs[:] # clone via list slicing - very fast
        return dirs

    # Returns true if the given tile has more than 2 free directions
    # (x,y) - Tile coordinates of the tile to check
    def isCrossing(self, (x,y)):
        freedirs = self.getFreeDirs((x,y))
        return len(freedirs) > 2

    # Returns True if the given tile is exactly 1 tile next to a crossing
    def isNextToCrossing(self, (x,y)):
        freedirs = self.getFreeDirs((x,y))
        for freedir in freedirs:
            freevec = freedir.tuple()
            neighbor = (x + freevec[0], y + freevec[1])
            if self.isCrossing(neighbor):
                return True

        return False

    # Returns true if this tile is part of the maze path
    # (x,y) - Tile coordinates of the tile to check
    def isPathTile(self, (x,y)):
        return (x,y) in self.__path

    # Returns the tile at tile-position (x,y)
    def getTile(self, (x,y)):
        if (x,y) in self.__tiles:
            return self.__tiles[(x,y)]
        else:
            return None

    # Returns the tile at pixel-position (x,y)
    def getTileAtPixel(self, (x,y)):
        return self.getTile(World.getTilePos((x,y)))

    # Generates random curve points in a fixed grid. Each grid cell contains exactly one
    # curve point with a 1-tile padding in each direction.
    def __generateRandomCurvePoints(self):
        curvePoints = [(1,1)] # (1,1) is the (separate) start point of the path

        xRand = 1
        yRand = 5 # give the start point space on the north side of the world
        for h in range(1, int(self.__height // self.__gridSize) + 1):
            for w in range(int(self.__width // self.__gridSize)):
                curvePoints.append((choice([xRand, xRand + 2]), choice([yRand, yRand + 2])))
                xRand += 5
            xRand = 1
            yRand += 5

        return curvePoints

    # Determine a path through a random set of curve points by finding the next nearest
    # curve point of the current curve point, skipping already visited points.
    def __calculateRandomCurvePointPath(self):
        notVisited = self.__generateRandomCurvePoints()
        pathCurvePoints = [notVisited[0]]
        currentNode = notVisited[0]
        nearestNode = notVisited[0]
        del notVisited[0]
        while len(notVisited) > 0:
            minDist = sys.maxint
            if len(notVisited) == 1:
                pathCurvePoints.append(notVisited[0])
                break
            for i in notVisited:
                distSQ = (i[0] - currentNode[0]) ** 2 + (i[1] - currentNode[1]) ** 2
                if distSQ < minDist:
                    minDist = distSQ
                    nearestNode = i
            pathCurvePoints.append(nearestNode)
            notVisited.remove(nearestNode)
            currentNode = nearestNode
        return pathCurvePoints

    # Returns list of tuples of each tile position in the path
    def __calculateRandomPath(self):
        path = []
        pathCurvePoints = self.__calculateRandomCurvePointPath()
        for i in xrange(len(pathCurvePoints) - 1):
            cur, nxt = pathCurvePoints[i], pathCurvePoints[i + 1]
            sgn = (1 if nxt[0] >= cur[0] else -1,
                   1 if nxt[1] >= cur[1] else -1)
            for x in xrange(cur[0], nxt[0] + sgn[0], sgn[0]):
                p = (x,cur[1])
                if not p in path:
                    path.append(p)
            for y in xrange(cur[1] + sgn[1], nxt[1] + sgn[1], sgn[1]):
                p = (nxt[0],y)
                if not p in path:
                    path.append(p)

        return path

    # Create the actual Tile instances for each tile position defined in randArray
    def __setPath(self, path, createRandomEvents=True):
        # Clear previous tiles
        # WARNING: This does NOT clear the features. It is assumed that the world
        #          is created from scratch after destroying it.
        for tilePos in self.__path:
            self.__tiles[tilePos].texture = ""
        self.__freeDirectionCache = {}

        # Copy path
        self.__path = list(path)

        # Update properties of path tiles
        eventDistanceCount = 0
        for pos in path:
            if not createRandomEvents:
                isEvent = False
            elif eventDistanceCount > self.__minEventDistance:
                if randint(0, 1) == 0:
                    isEvent = True
                    eventDistanceCount = 0
                else:
                    isEvent = False
                    eventDistanceCount += 1
            else:
                isEvent = False
                eventDistanceCount += 1

            freeDirs = self.getFreeDirs(pos)
            numFreeDirs = len(freeDirs)
            tileType = 'ground'
            tileDirs = [] # default directions, defined by the asset
            if numFreeDirs == 1:
                tileType = 'dead_end'
                tileDirs = [Direction.DOWN]
            elif numFreeDirs == 2:
                if abs(freeDirs[0].angle(freeDirs[1])) == 180:
                    tileType = 'path_straight'
                    tileDirs = [Direction.UP, Direction.DOWN]
                else:
                    tileType = 'path_curve'
                    tileDirs = [Direction.RIGHT, Direction.DOWN]
                    isEvent = False
            elif numFreeDirs == 3:
                tileDirs = [Direction.LEFT, Direction.DOWN, Direction.RIGHT]
                isEvent = False
            else:
                tileDirs = [Direction.LEFT, Direction.UP, Direction.RIGHT, Direction.DOWN]
                isEvent = False

            freeDirs = [d.tuple() for d in freeDirs]
            tileDirs = [d.tuple() for d in tileDirs]
            angle = matchVectors(tileDirs, freeDirs)

            tile = self.__tiles[pos]
            tile.tileset = self.__tileset
            tile.texture = tileType
            tile.isevent = isEvent
            tile.rotation = angle

    def getTileset(self):
        return self.__tileset
