from libavg import avg, player as avgPlayer
from Tile import *
from TileEvent import *
from Direction import *
import random
import math

class Avalanche(TileEvent):
    EVENT_NAME = "avalanche_event"
    NUM_PILES_RANGE = (4,9) # min, max (inclusive)
    MAX_PILE_RAND_OFFSET = (1.4,0.4) # maximum random offset (-x,-y) to (x,y) in tiles (float)
    MAX_PILE_RAND_SIZE_FACTOR = 0.4 # 1.0-x <= randSizeFactor <= 1.0+x
    PILE_MOVE_SPEED = 120.0 # px per sec
    PILE_MOVE_TIME = 0.6 # duration of the avalanche animation
    PILE_MELT_DISTANCE = 140.0 # px
    PILE_MELT_TIME = 1.5 # sec
    PILE_DRAG_FACTOR = 0.6 # slows down the dragging of a pile

    # tile - must be a straight path tile
    def __init__(self, tile):
        super(Avalanche, self).__init__(tile)

        # Determine direction vector of the straight path tile
        world = tile.getWorld()
        freeDirs = world.getFreeDirs(tile.position)
        self.__pathDir = freeDirs[0]
        self.__pathVec = avg.Point2D(self.__pathDir.tuple())

        # Choose between both vectors that are perpendicular to the path dir
        self.__perpDir = self.__pathDir.rotatedBy90CW()
        if random.choice([True, False]):
            self.__perpDir = self.__perpDir.inverted()

        self.__pathVecPerp = avg.Point2D(self.__perpDir.tuple())

        tileSz = avg.Point2D(Tile.getTileSize())
        self.__pileBasePos = (tile.pos + 0.5 * tileSz) + self.__pathVecPerp * tileSz * 0.6

        # Spawn snow piles
        self.__piles = []
        for i in xrange(random.randint(Avalanche.NUM_PILES_RANGE[0], Avalanche.NUM_PILES_RANGE[1])):
            self.__spawnSnowPile()

        self.__avalancheTime = Avalanche.PILE_MOVE_TIME
        self.__inputEnabled = False

        self.__numMeltedPiles = 0

        avgPlayer.getRootNode().subscribe(avg.Node.CURSOR_MOTION, self.__onCursorMotion)
        avgPlayer.getRootNode().subscribe(avg.Node.CURSOR_UP, self.__onCursorUp)

        avgPlayer.subscribe(avg.Player.KEY_DOWN, self.__onKeyDown)

    def getEventName(self):
        return Avalanche.EVENT_NAME

    def isFullscreenEvent(self):
        return False

    def requiresCountdownAfterCompletion(self):
        return False

    def __spawnSnowPile(self):
        randOffs = avg.Point2D(
            random.uniform(-Avalanche.MAX_PILE_RAND_OFFSET[0], Avalanche.MAX_PILE_RAND_OFFSET[0]),
            random.uniform(-Avalanche.MAX_PILE_RAND_OFFSET[1], Avalanche.MAX_PILE_RAND_OFFSET[1]))

        # Swap x and y if vertically
        if self.__perpDir in [Direction.LEFT, Direction.RIGHT]:
            randOffs.x,randOffs.y = randOffs.y,randOffs.x

        tileSz = avg.Point2D(Tile.getTileSize())
        randPos = self.__pileBasePos + randOffs * tileSz
        randSzFactor = random.uniform(1.0 - Avalanche.MAX_PILE_RAND_SIZE_FACTOR, 1.0 + Avalanche.MAX_PILE_RAND_SIZE_FACTOR)

        tile = self.getTile()

        pile = avg.ImageNode(
            href="../assets/avalanche.png",
            pos=randPos,
            parent=tile.parent)
        pile.size *= randSzFactor
        pile.pos -= 0.5 * pile.size
        pile.angle = math.radians(tile.rotation + random.choice([0,180]))
        #pile.elementoutlinecolor = "FF0000"

        pile.melted = False

        pile.spawnPos = avg.Point2D(pile.pos)
        pile.lastCursorPos = None # Point2D (event.pos)
        pile.contactId = None
        pile.dragVec = avg.Point2D(0,0)
        pile.subscribe(avg.Node.CURSOR_DOWN, lambda e,pile=pile: self.__onPileCursorDown(e, pile))

        self.__piles.append(pile)
        return pile

    def onInit(self, parentDiv):
        # Spawn more piles
        for i in xrange(random.randint(Avalanche.NUM_PILES_RANGE[0], Avalanche.NUM_PILES_RANGE[1] * 2)):
            pile = self.__spawnSnowPile()
            pile.opacity = 0.0
            pile.anim = avg.LinearAnim(pile, "opacity", int(Avalanche.PILE_MOVE_TIME * 0.4 * 1000.0), 0.0, 1.0)
            pile.anim.start()

    def onFrame(self, dt, keystates):
        dt *= 0.001 # ms to sec

        if self.__avalancheTime > 0.0:
            self.__avalancheTime -= dt

            for pile in self.__piles:
                # Move pile towards the rail
                pile.pos -= self.__pathVecPerp * Avalanche.PILE_MOVE_SPEED * dt

                # Make tile bigger
                pile.size *= 1.0 + 0.35 * dt
        else:
            self.__inputEnabled = True

            for pile in self.__piles:
                if pile.melted and pile.opacity > 0.0:
                    pile.pos += pile.dragVec * Avalanche.PILE_DRAG_FACTOR
                    pile.dragVec *= 0.7
                    pile.opacity -= (1.0 / Avalanche.PILE_MELT_TIME) * dt

        if self.__numMeltedPiles >= len(self.__piles):
            anims = []
            for pile in self.__piles:
                if pile.opacity > 0.0:
                    anims.append(avg.LinearAnim(pile, "opacity", 300, pile.opacity, 0.0))
            parallelAnim = avg.ParallelAnim(anims)
            parallelAnim.start()
            return True
        else:
            return False

    def __onPileCursorDown(self, event, pile):
        if not self.__inputEnabled:
            return
        if not pile.melted and pile.contactId is None:
            pile.contactId = event.cursorid
            pile.lastCursorPos = event.pos

    def __onCursorMotion(self, event):
        if event.contact is None:
            return

        for pile in self.__piles:
            if pile.contactId == event.cursorid:
                pile.dragVec = event.pos - pile.lastCursorPos
                pile.pos += pile.dragVec * Avalanche.PILE_DRAG_FACTOR
                pile.lastCursorPos = event.pos
                tileSz = avg.Point2D(Tile.getTileSize())
                if ((pile.pos + 0.5 * pile.size) - (self.getTile().pos + 0.5 * tileSz)).getNorm() > Avalanche.PILE_MELT_DISTANCE:
                    pile.contactId = None
                    pile.melted = True
                    pile.sensitive = False
                    self.__numMeltedPiles += 1
                break

    def __onCursorUp(self, event):
        for pile in self.__piles:
            if pile.contactId == event.cursorid:
                pile.contactId = None
                break

    def __onKeyDown(self, event):
        if event.keyname == 'S':
            for pile in self.__piles:
                pile.melted = True
                pile.contactId = None
            self.__inputEnabled = False
            self.__numMeltedPiles = len(self.__piles)
