from libavg import avg, player as avgPlayer
import random
import math
from Util import *
from TileEvent import TileEvent


class BridgeEvent(TileEvent):
    EVENT_NAME = "bridge_event"
    FALL_ANIM_DURATION = 1000 # ms

    def getEventName(self):
        return self.EVENT_NAME

    def isFullscreenEvent(self):
        return True

    # difficulty - from 0 (no holes) to 1 (very hard)
    def __init__(self, tile, difficulty=0.8):
        super(BridgeEvent, self).__init__(tile)
        self.difficulty = difficulty
        self.__moveUp = False
        self.__moveDown = False
        self.__bridge = avg.ImageNode(pos=tile.pos, href="../assets/bridge.png", angle=math.radians(tile.rotation), size=tile.getTileSize(), parent=tile.parent)


    def onInit(self, parentDiv):
        scale = 550.0 / 800.0
        numTiles = 14
        tileSz = (200 * scale, 1400 * scale)
        holeSz = (200 * scale, 200 * scale)
        self.__borderWidth = 130 * scale
        self.__innerBridgeHeight = tileSz[1] - 2 * self.__borderWidth
        holeOffsetRange = self.__innerBridgeHeight - holeSz[1]
        pawnSz = (103 * scale, 179 * scale)
        self.__bridgeWidth = numTiles * tileSz[0]
        self.__gameOffset = avg.Point2D((parentDiv.width - self.__bridgeWidth) * 0.5, (parentDiv.height - tileSz[1]) * 0.2)
        self.__pawnSpeed = (0.2, 0.4)

        self.__blackboxOffset = 250 + (1.0 - self.difficulty) * 250.0

        holeDist = numTiles * 0.1 # horizontally

        holeYOffs = holeOffsetRange * 1.5 * self.difficulty

        avg.ImageNode(href="../assets/bridge-event/bridge-tile.png", pos=parentDiv.pos, parent=parentDiv)

        self.__holes = []
        lastHole = 0
        yOffs = holeOffsetRange * 0.5
        for i in range(0, numTiles):
            pos = avg.Point2D(self.__gameOffset.x + tileSz[0] * i, self.__gameOffset.y)

            # Place a hole randomly:
            if i - lastHole >= holeDist:
                yOffs = (yOffs + random.random() * holeYOffs) % holeOffsetRange
                holePos = avg.Point2D(pos.x + (tileSz[0] - holeSz[0]) * 0.5, pos.y + self.__borderWidth + yOffs)

                angles = [0, 90, 180, 270]
                hole = avg.ImageNode(href="../assets/bridge-event/bridge-hole.png", pos=holePos, size=holeSz, angle=math.radians(random.choice(angles)), parent=parentDiv)
                self.__holes.append(hole)
                lastHole = i

        self.__pawnSpawnPos = avg.Point2D(self.__gameOffset.x, self.__gameOffset.y + tileSz[1] * 0.5)
        self.__pawnSpawnSize = pawnSz
        self.__pawn = avg.ImageNode(href="../assets/bridge-event/pawn.png", pos=self.__pawnSpawnPos, size=self.__pawnSpawnSize, parent=parentDiv)

        self.__blackbox = avg.RectNode(pos=(self.__pawnSpawnPos.x + self.__blackboxOffset, parentDiv.y), size=(parentDiv.width, self.__bridgeWidth), color="000000", fillcolor="000000", fillopacity=1.0, parent=parentDiv)

        # Create buttons:
        btnScreenDist = 20 # Margin from screen border to buttons
        btnSz = avg.Point2D(181, 181) * scale
        btnUp = avg.ImageNode(href="../assets/bridge-event/arrow.png", pos=((parentDiv.width - btnSz.x) * 0.5, btnScreenDist), size=btnSz, parent=parentDiv)
        btnUp.subscribe(btnUp.CURSOR_DOWN, self.__onBtnUp)
        btnUp.subscribe(btnUp.CURSOR_UP, self.__onBtnUpRelease)

        btnDown = avg.ImageNode(href="../assets/bridge-event/arrow.png", pos=((parentDiv.width - btnSz.x) * 0.5, parentDiv.height - btnScreenDist - btnSz.x), size=btnSz, angle=3.14, parent=parentDiv)
        btnDown.subscribe(btnDown.CURSOR_DOWN, self.__onBtnDown)
        btnDown.subscribe(btnDown.CURSOR_UP, self.__onBtnDownRelease)

        avgPlayer.subscribe(avgPlayer.KEY_UP, self.__onKeyUp)
        self.__skip = False
        self.__fallAnimation = None

    def __onBtnUp(self, event):
        self.__moveUp = True

    def __onBtnUpRelease(self, event):
        self.__moveUp = False

    def __onBtnDown(self, event):
        self.__moveDown = True

    def __onBtnDownRelease(self, event):
        self.__moveDown = False

    def __onKeyUp(self, event):
        if event.keyname == 'S':
            self.__skip = True

    def onDisabled(self):
        self.enable()

    def __onFallAnimationFinished(self):
        self.__pawn.pos = self.__pawnSpawnPos
        self.__pawn.size = self.__pawnSpawnSize
        self.__pawn.angle = 0
        self.__blackbox.pos = avg.Point2D(self.__pawnSpawnPos.x + self.__blackboxOffset, 0)
        self.__fallAnimation = None

    def onFrame(self, dt, keystates):
        if self.__skip:
            if self.__fallAnimation is not None:
                self.__fallAnimation.abort()
                self.__fallAnimation = None

            return True

        if (self.__pawn.x + self.__pawn.width * 0.5) >= (self.__gameOffset.x + self.__bridgeWidth):
            return True

        if self.__fallAnimation is not None:
            return False

        # Collision detection
        pawnAABB = AABB(self.__pawn.pos, self.__pawn.pos + self.__pawn.size)
        intersectedHole = None
        for hole in self.__holes:
            holeAABB = AABB(hole.pos, hole.pos + hole.size)
            holeAABB.inset(avg.Point2D(30,30))
            if pawnAABB.intersects(holeAABB):
                intersectedHole = hole
                break

        # When hitting a hole, show fall animation and restart the game
        if intersectedHole is not None:
            holeCenter = intersectedHole.pos + 0.5 * intersectedHole.size
            self.__fallAnimation = avg.ParallelAnim([
                avg.LinearAnim(self.__pawn, "size", BridgeEvent.FALL_ANIM_DURATION, self.__pawn.size, avg.Point2D(1,1)),
                avg.LinearAnim(self.__pawn, "pos", BridgeEvent.FALL_ANIM_DURATION, self.__pawn.pos, holeCenter),
                avg.LinearAnim(self.__pawn, "angle", BridgeEvent.FALL_ANIM_DURATION, self.__pawn.angle, 14.0)
            ], None, self.__onFallAnimationFinished)
            self.__fallAnimation.start()
            return

        # Move the game forward
        self.__pawn.x += self.__pawnSpeed[0] * dt
        self.__blackbox.pos = self.__blackbox.pos + avg.Point2D(self.__pawnSpeed[0] * dt, 0)

        ySpeed = 0.0
        if keystates['Up'] or self.__moveUp:
            ySpeed += -self.__pawnSpeed[1]
        if keystates['Down'] or self.__moveDown:
            ySpeed += self.__pawnSpeed[1]

        y = self.__pawn.y + ySpeed * dt
        if y > (self.__gameOffset.y + self.__borderWidth) and (y + self.__pawn.height) < (self.__gameOffset.y + self.__borderWidth + self.__innerBridgeHeight):
            self.__pawn.y = y
