from libavg import avg, player as avgPlayer
from Tile import *
from TileEvent import *
from random import randint
from Util import AABB
from Direction import Direction
import math

# Feature to mark a tile as a part of the BrokenSwampRails event
class BrokenSwampRailFeature(TileFeature):
    FEATURE_NAME = "broken_swamp_rail_feature"

    def __init__(self, tile, rootEvent):
        super(BrokenSwampRailFeature, self).__init__(tile)
        self.__rootEvent = rootEvent

        if tile is not rootEvent.getTile():
            tile.addFeature(rootEvent)

    def onDestroy(self):
        if self.__rootEvent:
            self.getTile().removeFeature(self.__rootEvent.getFeatureName(), False)
            self.__rootEvent.removeFeature(self)
            self.__rootEvent = None

    def getFeatureName(self):
        return BrokenSwampRailFeature.FEATURE_NAME


class BrokenSwampRails(TileEvent):
    EVENT_NAME = "broken_swamp_rails_event"
    DRIFT_OFF_AMOUNT = 10.0 # px per sec into an arbitrary direction
    DRIFT_OFF_TIME = 0.8 # duration of drift-off in seconds
    DRIFT_OFF_MAX_ANGLE = 30 # degrees, maximum variation of the orthogonal player direction
    BOUND_BOX_OUTSET = 2 # no of tiles, the bound box is expanded in each direction
    DRAG_SPEED = 50.0 # px per sec

    def __init__(self, rootTile):
        super(BrokenSwampRails, self).__init__(rootTile)
        self.__features = [] # features of affected path tiles
        self.__stuck = False
        self.__player = None
        self.__playerInitSpeed = 0 # speed when event started
        self.__playerDriftSpeed = 0 # speed that is decelerated until reaching 0
        self.__driftOffDir = None # Point2D, playerDir vector rotated a little bit
        self.__playerOrigSensitive = None
        self.__contactId = None # Drag-contact
        self.__lastCursorPos = None
        self.__aabb = AABB()

        # Distribute broken-rails asset on nearby connected path tiles
        world = rootTile.getWorld()
        rootPos = rootTile.position
        radiusSq = randint(1, 3) ** 2
        open, closed = [rootPos], []
        while len(open) > 0:
            pos = open.pop()
            closed.append(pos)
            tile = world.getTile(pos)
            tile.removeAllFeatures() # override existing features on this tile
            tile.opacity = 0.0

            feature = BrokenSwampRailFeature(tile, self)
            tile.addFeature(feature)
            self.__features.append(feature)

            freeDirs = world.getFreeDirs(pos)
            goodFreeDirs = 0
            for freeDir in freeDirs:
                freeDirVec = freeDir.tuple()
                freePos = (pos[0] + freeDirVec[0], pos[1] + freeDirVec[1])
                if ((freePos[0] - rootPos[0]) ** 2 + (freePos[1] - rootPos[1]) ** 2) > radiusSq:
                    continue
                if not world.isPathTile(freePos) or freePos in closed:
                    continue
                freeTile = world.getTile(freePos)
                if freeTile.hasFeature(LootChest.FEATURE_NAME):
                    continue
                open.append(freePos)
                goodFreeDirs += 1

            if goodFreeDirs == 0:
                tile.opacity = 1.0
                tile.texture = "broken_swamp_rail"

                if len(freeDirs) <= 2:
                    for freeDir in freeDirs:
                        freeDirVec = freeDir.tuple()
                        freePos = (pos[0] + freeDirVec[0], pos[1] + freeDirVec[1])
                        if not freePos in closed:
                            # Let the asset point into this direction
                            tile.angle = Direction.DOWN.angleRadians(freeDir)
                            break


        self.__updateFeatureBoundBox()
        rootTile.parent.subscribe(avg.Node.CURSOR_UP, self.__onCursorUp)

    def getEventName(self):
        return BrokenSwampRails.EVENT_NAME

    def isDisabled(self):
        # This event cannot be disabled
        return False

    def onDestroy(self):
        for feature in self.__features:
            feature.destroy()
            featureTile = feature.getTile()
            if featureTile:
                featureTile.removeFeature(feature.getFeatureName())
        self.__features = []

    def removeFeature(self, feature):
        if feature in self.__features:
            self.__features.remove(feature)
        self.__updateFeatureBoundBox()

    def isFullscreenEvent(self):
        # Do not go into fullscreen mode but prevent regular player simulation
        return False

    def requiresCountdownAfterCompletion(self):
        return False

    def __updateFeatureBoundBox(self):
        self.__aabb = AABB()
        for feature in self.__features:
            featureTile = feature.getTile()
            if featureTile:
                self.__aabb.add(featureTile.pos)
                self.__aabb.add(featureTile.pos + avg.Point2D(Tile.getTileSize()))
        outset = avg.Point2D(BrokenSwampRails.BOUND_BOX_OUTSET,BrokenSwampRails.BOUND_BOX_OUTSET)
        self.__aabb.outset(outset * avg.Point2D(Tile.getTileSize()))

    # Called once everytime the player reaches this event
    def onInit(self, parentDiv, player):
        self.__stuck = True
        if self.__player is not player:
            if self.__player:
                self.__player.unsubscribe(avg.Node.CURSOR_DOWN, self.__onPlayerCursorDown)
                self.__player.unsubscribe(avg.Node.CURSOR_MOTION, self.__onPlayerCursorMotion)

            self.__player = player
            self.__player.subscribe(avg.Node.CURSOR_DOWN, self.__onPlayerCursorDown)
            self.__player.subscribe(avg.Node.CURSOR_MOTION, self.__onPlayerCursorMotion)

        self.__playerOrigSensitive = player.sensitive
        self.__player.sensitive = True
        self.__playerDriftSpeed = player.speed
        self.__playerInitSpeed = player.speed

        driftOffAngle = randint(-BrokenSwampRails.DRIFT_OFF_MAX_ANGLE, BrokenSwampRails.DRIFT_OFF_MAX_ANGLE)
        self.__driftOffDir = avg.Point2D(player.orientation.tuple()).getRotated(math.radians(driftOffAngle))

    def __onCompleted(self):
        self.__player.sensitive = self.__playerOrigSensitive

    # Checks if player is approx. on an actual path tile now.
    # Returns True if completed, False if not
    def __checkIfCompleted(self):
        if not self.__stuck:
            return True

        if not self.__player:
            return False

        world = self.getTile().getWorld()
        pxpos = self.__player.pos
        tilesz = Tile.getTileSize()
        approxTilePos = (round(pxpos.x / tilesz[0]), round(pxpos.y / tilesz[1]))
        if not world.isPathTile(approxTilePos) or world.isCrossing(approxTilePos):
            return False

        tile = world.getTile(approxTilePos)
        if tile.hasFeature(BrokenSwampRailFeature.FEATURE_NAME):
            return False

        # completed!
        self.__stuck = False
        self.__player.pos = world.getPixelPos(approxTilePos)
        self.__onCompleted()
        return True

    def __onPlayerCursorDown(self, event):
        if self.__stuck and self.__contactId is None:
            self.__contactId = event.cursorid
            self.__lastCursorPos = event.pos

    def __onCursorUp(self, event):
        if self.__contactId is not None and self.__contactId == event.cursorid:
            self.__contactId = None


    def __onPlayerCursorMotion(self, event):
        if self.__contactId is not None and self.__contactId == event.cursorid:
            dt = avgPlayer.getFrameDuration() * 0.001
            diff = event.pos - self.__lastCursorPos
            self.__lastCursorPos = event.pos

            newpos = self.__player.pos + diff * dt * BrokenSwampRails.DRAG_SPEED
            tilePxSz = avg.Point2D(Tile.getTileSize())
            worldPxSz = avg.Point2D(self.getTile().getWorld().getSize()) * tilePxSz
            if (0 < newpos.x < worldPxSz.x) and (0 < newpos.y < worldPxSz.y):
                if self.__aabb.contains(newpos) and self.__aabb.contains(newpos + tilePxSz):
                    self.__player.pos = newpos

            self.__checkIfCompleted()


    # Returns True when completed, False if not
    def onFrame(self, dt, keystats):
        if self.__stuck:
            # Apply player drift-off
            newPlayerPos = self.__player.pos + self.__driftOffDir * dt * 0.001 * self.__playerDriftSpeed
            worldPxSz = avg.Point2D(self.getTile().getWorld().getSize()) * avg.Point2D(Tile.getTileSize())
            if (0 < newPlayerPos.x < worldPxSz.x) and (0 < newPlayerPos.y < worldPxSz.y):
                self.__player.pos = newPlayerPos

            # Apply deceleration
            self.__playerDriftSpeed -= (self.__playerInitSpeed / BrokenSwampRails.DRIFT_OFF_TIME) * dt * 0.001
            if self.__playerDriftSpeed < 0.0:
                self.__playerDriftSpeed = 0

        return not self.__stuck
