from libavg import avg, player as avgPlayer
from TileEvent import *
from Util import *
import math
import random

class LeverRiddleEvent(TileEvent):
    EVENT_NAME = "leverriddle_event"
    TURNED_ON_LIGHT_COLOR = "ffeda4"
    ROW_HIGHLIGHT_COLOR = "332222"
    HINT_COLOR = "111111"
    HINT_OPACITY = 0.7
    HINT_OPACITY_DECAY = 0.35 # per row
    TIME_TO_COMPLETE = 20 # sec

    def __init__(self, tile):
        super(LeverRiddleEvent, self).__init__(tile)
        self.__wallImg = avg.ImageNode(
            href="../assets/leverriddle-wall.png",
            pos=tile.pos,
            size=tile.size,
            angle=math.radians(tile.rotation),
            parent=tile.parent)

        self.__completed = False
        self.__blockInput = False

    def getEventName(self):
        return LeverRiddleEvent.EVENT_NAME

    def isFullscreenEvent(self):
        return True

    def __generateRandomFormula(self):
        contributions = {}
        levers = range(0, 4)
        random.shuffle(levers)
        remaining = 4
        while len(levers) > 1:
            amount = random.randint(0, random.choice([remaining // 2, remaining]))
            contributions[levers.pop(0)] = amount
            remaining -= amount

        contributions[levers.pop(0)] = remaining

        for c in contributions:
            if contributions[c] == 0:
                contributions[c] = -random.randint(0,10)

        return contributions

    def onInit(self, parentDiv):
        # Show lamps in the center - they are partially hidden behind the background
        numLamps = 4
        numRows = 3
        lampRadius = 50 # px
        lampDist = (105,105) # px
        screenCenter = parentDiv.size * 0.5
        self.__lamps = []
        for row in xrange(numRows):
            lamps = []
            yoffs = (row + 1 - 0.5 * numRows) * (lampDist[1] + 2.0 * lampRadius) - 0.5 * lampDist[1] - lampRadius

            ids = range(0, numLamps)
            random.shuffle(ids)
            for i in ids:
                # xoffs = n(d + 2r) - d/2 - r     where n from -N/2 to N/2
                xoffs = (i + 1 - 0.5 * numLamps) * (lampDist[0] + 2.0 * lampRadius) - 0.5 * lampDist[0] - lampRadius
                lamp = avg.CircleNode(
                    pos=screenCenter + avg.Point2D(xoffs,yoffs), r=lampRadius,
                    color="555555", strokewidth=0,
                    fillcolor="000000", fillopacity=1.0,
                    parent=parentDiv)

                if row == 0:
                    lamp.color = LeverRiddleEvent.ROW_HIGHLIGHT_COLOR

                lamps.append(lamp)

            self.__lamps.append(lamps)

        # Background
        self.__background = avg.ImageNode(href="../assets/lever-riddle/background.png", size=parentDiv.size, parent=parentDiv)

        # Show levers in each corner
        self.__levers = []
        self.__hints = []
        mainPadding = avg.Point2D(195,180) # px
        rightCorrectionPadding = 20 # px - due to perspective of the handle assets
        leverSz = avg.Point2D(173,263) # px
        handleSz = avg.Point2D(195,65) # px
        leverAssetNames = ['fader_top_left.png', 'fader_top_right.png', 'fader_bottom_left.png', 'fader_bottom_right.png']
        for i in xrange(4):
            pos = mainPadding + avg.Point2D(i % 2, i // 2) * (parentDiv.size - 2.0 * mainPadding - leverSz)
            isright = (i % 2 == 1)
            if isright:
                pos.x -= rightCorrectionPadding

            lever = avg.DivNode(pos=pos, size=leverSz, parent=parentDiv)
            lever.leverHandle = avg.ImageNode(href="../assets/lever-riddle/"+leverAssetNames[i], size=handleSz, parent=lever)
            self.__levers.append(lever)

            lever.isUpsideDown = (i < 2) # first ones are on the top of the screen
            lever.aabb = AABB(lever.pos, lever.pos + lever.size)

            lever.cursorId = None
            lever.subscribe(avg.Node.CURSOR_MOTION, lambda e,lever=lever: self.__onLeverCursorMotion(e, lever))
            lever.leverHandle.subscribe(avg.Node.CURSOR_DOWN, lambda e,lever=lever: self.__onLeverHandleCursorDown(e, lever))

        # Show timer bar
        timerOffsX = ((0 + 1 - 0.5 * numLamps) * (lampDist[0] + 2.0 * lampRadius) - 0.5 * lampDist[0] - lampRadius) - lampRadius
        timerOffsY = (numRows + 1 - 0.5 * numRows) * (lampDist[1] + 2.0 * lampRadius) - 0.5 * lampDist[1] - lampRadius
        timerOffsY -= 100
        timerWidth = self.__timerWidth = numLamps * (2 * lampRadius) + (numLamps - 1) * lampDist[0]
        timerHeight = 15
        timerPos1 = screenCenter + avg.Point2D(timerOffsX,timerOffsY)
        self.__timer = avg.LineNode(
            pos1=timerPos1, pos2=timerPos1 + avg.Point2D(timerWidth,0),
            color="FF0000",
            strokewidth=timerHeight,
            parent=parentDiv)

        self.__remainingTime = 0

        avgPlayer.subscribe(avgPlayer.KEY_UP, self.__onKeyUp)
        parentDiv.subscribe(avg.Node.CURSOR_UP, self.__onCursorUp)
        parentDiv.subscribe(avg.Node.CURSOR_MOTION, self.__onCursorMotion)

        self.__restartGame()

    def __restartGame(self):
        # Generate three magic formulas
        self.__contributions = [self.__generateRandomFormula() for i in xrange(len(self.__lamps))]

        self.__currentRow = 0
        self.__lastval = -1

        # Reset lamps
        for row in xrange(len(self.__lamps)):
            for lamp in self.__lamps[row]:
                if row == self.__currentRow:
                    lamp.fillcolor = LeverRiddleEvent.ROW_HIGHLIGHT_COLOR
                else:
                    lamp.fillcolor = "000000"

        # Reset levers
        self.__resetLevers()

        # Decay hints
        garbageHints = []
        for hint in self.__hints:
            hint.opacity -= LeverRiddleEvent.HINT_OPACITY_DECAY * 1.5
            if hint.opacity <= 0:
                hint.unlink(True)
                garbageHints.append(hint)

        for hint in garbageHints:
            self.__hints.remove(hint)

        # Distribute hints for the first row
        self.__distributeHints()

        # Reset timer
        self.__remainingTime = LeverRiddleEvent.TIME_TO_COMPLETE

    def onFrame(self, dt, keystates):
        self.__updateTimer()

        for lever in self.__levers:
            if not lever.cursorId:
                # Move the handle more to the side of the levers current state
                moveSpeed = 200 # px per sec
                state = self.__getLeverState(lever)
                ydir = (-1 if state else 1)
                if lever.isUpsideDown:
                    ydir *= -1.0

                self.__moveLeverHandle(lever, ydir * dt * 0.001 * moveSpeed)

        self.__updateLights()

        if self.__completed:
            return True
        else:
            self.__remainingTime -= dt * 0.001
            if self.__remainingTime <= 0:
                self.__restartGame()
            return False

    def __onKeyUp(self, event):
        if event.keyname == 'S':
            self.__currentRow = len(self.__lamps)
            self.__updateTimer()

            for lampRow in self.__lamps:
                for lamp in lampRow:
                    lamp.fillcolor = LeverRiddleEvent.TURNED_ON_LIGHT_COLOR

            self.__onCompleted()

    def __onLeverHandleCursorDown(self, event, lever):
        if self.__blockInput:
            return

        if not lever.cursorId:
            lever.cursorId = event.cursorid
            lever.cursorLastPos = event.pos

    def __onLeverCursorMotion(self, event, lever):
        if self.__blockInput:
            return

        if lever.cursorId and event.cursorid == lever.cursorId:
            # Project motion onto y-axis
            yaxis = avg.Point2D(0,1)
            motion = yaxis * (event.pos - lever.cursorLastPos).dot(yaxis)
            self.__moveLeverHandle(lever, motion.y)
            lever.cursorLastPos = event.pos

    # parentDiv's CURSOR_MOTION
    def __onCursorMotion(self, event):
        if self.__blockInput:
            return

        for lever in self.__levers:
            if lever.cursorId == event.cursorid:
                if not lever.aabb.contains(event.pos):
                    lever.cursorId = None
                    return

    def __onCursorUp(self, event):
        if self.__blockInput:
            return

        for lever in self.__levers:
            if lever.cursorId == event.cursorid:
                lever.cursorId = None
                break

    # Returns True if the lever is in "enabled" state, False if not
    def __getLeverState(self, lever):
        handle = lever.leverHandle
        handlepos = (handle.pos.y + 0.5 * handle.size.y) - 0.5 * lever.size.y
        if lever.isUpsideDown:
            handlepos *= -1.0
        return handlepos < 0.0

    def __moveLeverHandle(self, lever, ydiff):
        handle = lever.leverHandle
        newy = handle.pos.y + ydiff
        newy = max(min(newy, lever.height - handle.size.y), 0)
        handle.pos = avg.Point2D(handle.pos.x, newy)

    # Resets all levers to their OFF-position
    def __resetLevers(self):
        for lever in self.__levers:
            handle = lever.leverHandle
            handley = (0 if lever.isUpsideDown else lever.height - handle.size.y)
            handle.pos = avg.Point2D(handle.pos.x, handley)
            lever.cursorId = None

    def __updateLights(self):
        if self.__currentRow == len(self.__lamps):
            return

        # Determine value of current combination
        val = 0
        for i in xrange(4):
            leverState = self.__getLeverState(self.__levers[i])
            val += int(leverState) * self.__contributions[self.__currentRow][i]

        val = max(0, val)

        if val == self.__lastval:
            return

        self.__lastval = val

        # Turn on lamps based on current combination
        lampCount = val
        row = self.__lamps[self.__currentRow]
        for i in xrange(len(row)):
            row[i].fillcolor = (LeverRiddleEvent.TURNED_ON_LIGHT_COLOR if i < lampCount else LeverRiddleEvent.ROW_HIGHLIGHT_COLOR)

        # Determine if row is completed
        if val == len(self.__lamps[self.__currentRow]):
            self.__onRowCompleted()

    def __updateTimer(self):
        if self.__currentRow == len(self.__lamps):
            self.__timer.active = False
            return

        width = self.__timerWidth * (self.__remainingTime / LeverRiddleEvent.TIME_TO_COMPLETE)
        self.__timer.pos2 = self.__timer.pos1 + avg.Point2D(width,0)

    def __onRowCompleted(self):
        self.__currentRow += 1

        self.__resetLevers()

        numRows = len(self.__lamps)
        if self.__currentRow == numRows:
            self.__onCompleted()
        else:
            for i in xrange(numRows):
                for lamp in self.__lamps[i]:
                    if i == self.__currentRow:
                        lamp.fillcolor = LeverRiddleEvent.ROW_HIGHLIGHT_COLOR

            self.__distributeHints()

    # Distributes the hint numbers on the levers for the current row
    def __distributeHints(self):
        if self.__currentRow >= len(self.__lamps):
            return

        # Reduce opacity of all existing hints as they "age"
        for hint in self.__hints:
            hint.opacity -= LeverRiddleEvent.HINT_OPACITY_DECAY

        currentRowContrib = self.__contributions[self.__currentRow]
        padding = ((45,50),(40,70)) # ((left,top),(right,bottom)) - padding inside the "lever"
        for i in xrange(len(currentRowContrib)):
            lever = self.__levers[i]
            contrib = currentRowContrib[i]

            # Do not show a hint if contribution of the lever is 0
            if contrib == 0:
                continue

            fontsize = random.randint(18, 50) #px
            angle = (180 if lever.isUpsideDown else 0) + random.randint(-30,30)
            pos = avg.Point2D(
                random.uniform(padding[0][0], lever.width - padding[1][0]),
                random.uniform(padding[0][1], lever.height - padding[1][1]))

            # Add hint
            hint = avg.WordsNode(
                text=`contrib`,
                pos=pos,
                angle=math.radians(angle),
                font="Gloria Hallelujah",
                fontsize=fontsize,
                color=LeverRiddleEvent.HINT_COLOR,
                opacity=LeverRiddleEvent.HINT_OPACITY,
                parent=lever)

            extents = hint.getLineExtents(0)
            if hint.x + extents.x > (lever.width - padding[1][0]):
                hint.pos = avg.Point2D(lever.width - padding[1][0] - extents.x, hint.y)

            self.__hints.append(hint)

            bringNodeToTop(lever.leverHandle)


    # This is just a stupid workaround because python does not allow
    # simple variable assignment in lambdas...
    def __onReallyCompleted(self):
        self.__completed = True
        self.disable()
        self.__wallImg.active = False

    def __onCompleted(self):
        self.__blockInput = True

        # Highlight all rows green
        for lampRow in self.__lamps:
            for lamp in lampRow:
                lamp.fillcolor = "558811"

        avg.Player.get().setTimeout(2000, self.__onReallyCompleted)
