from libavg import avg, player as avgPlayer
from TileEvent import *
from World import World
import random
import math

class WaterPuddle(TileFeature):
    FEATURE_NAME = "waterpuddle_feature"

    # eventRoot - the TileEvent which is the root of this water puddle
    def __init__(self, tile, eventRoot):
        super(WaterPuddle, self).__init__(tile)
        self.__eventRoot = eventRoot

        if tile is not eventRoot.getTile():
            tile.addFeature(eventRoot)

        # Create the tile overlay with an opacity based on the distance
        # to the event root
        worldRootNode = tile.parent
        dist = (eventRoot.getTile().pos - tile.pos).getNorm() / Tile.getTileSize()[0]
        maxDist = 5.0
        opacity = 1.0 - min(0.2 + dist * (1.0 / maxDist), 0.6)
        self.__image = avg.ImageNode(
            href="../assets/" + tile.getTileset() + "/path_straight_flooded.png",
            pos=tile.pos,
            size=tile.size,
            angle=math.radians(tile.rotation),
            opacity=opacity,
            parent=worldRootNode)
        self.__image.sensitive = False

    def getFeatureName(self):
        return WaterPuddle.FEATURE_NAME

    def destroy(self):
        if self.__image:
            self.__image.unlink()
        self.__image = None
        self.getTile().removeFeature(TileEvent.FEATURE_NAME)
        self.__eventRoot = None


class PipesEvent(TileEvent):
    EVENT_NAME = "flooding_event"

    def __init__(self, rootTile):
        super(PipesEvent, self).__init__(rootTile)

        self.__completed = False

        # Decide difficulty of Event at creation not when it's triggered
        numY = random.randint(5,7)
        numX = numY + random.randint(1,2)
        self.__numTiles = (numX, numY)

        # Create nearby water tile features:
        self.__puddles = [] # WaterPuddle instances created by the algorithm
        world = rootTile.getWorld()
        worldRootNode = rootTile.parent
        open = [] # list of TILES we need to visit
        closed = [] # list of TUPLES of tile-positions that are marked as water puddles
        current = rootTile # current TILE we are visiting
        maxDistance = 3.0 # maximum distance from root tile in tile space
        while True:
            puddle = WaterPuddle(current, self)
            current.addFeature(puddle)
            self.__puddles.append(puddle)
            closed.append(current.position)

            freeDirs = world.getFreeDirs(current.position)
            for dir in freeDirs:
                dirTuple = dir.tuple()
                npos = (current.position[0] + dirTuple[0], current.position[1] + dirTuple[1])
                if (avg.Point2D(rootTile.position) - avg.Point2D(npos)).getNorm() <= maxDistance:
                    if not npos in closed:
                        tile = world.getTile(npos)
                        if tile and tile.getTexture() == 'path_straight' and not world.isCrossing(npos):
                            if not tile.hasFeature(TileEvent.FEATURE_NAME):
                                open.append(tile)

            if len(open) > 0:
                current = open.pop()
            else:
                break

        # Show broken pipe
        pipePos = rootTile.pos
        if len(closed) > 1:
            dirVec = avg.Point2D(0, -Tile.getTileSize()[1]).getRotated(math.radians(rootTile.rotation)) # direction defined by straight path asset
            secondTile = world.getTile(closed[1])
            pipePos += 1.5 * dirVec + 0.5 * (secondTile.pos - rootTile.pos)

        self.__brokenPipe = avg.ImageNode(
            href="../assets/pipe_broken.png",
            pos=pipePos,
            pivot=(0.5,0.5),
            angle=math.radians(rootTile.rotation),
            parent=worldRootNode)

        self.__brokenPipe.size = (Tile.getTileSize()[0], (Tile.getTileSize()[0] / self.__brokenPipe.width) * self.__brokenPipe.height)
        self.__brokenPipe.sensitive = False

    def getEventName(self):
        return PipesEvent.EVENT_NAME

    def isFullscreenEvent(self):
        return True

    def onDisabled(self):
        for puddle in self.__puddles:
            puddle.destroy()
        self.__brokenPipe.href = "../assets/pipe_fixed.png"

    def onInit(self, parentDiv):
        scale = (130.0 / 200.0, 130.0 / 200.0)
        tileSz = (200 * scale[0], 200 * scale[1]) #px
        gameSz = avg.Point2D(self.__numTiles[0] * tileSz[0], self.__numTiles[1] * tileSz[1])

        self.__assets = {
            'straight': "../assets/pipes-event/pipe-straight.png",
            'straight-empty': "../assets/pipes-event/pipe-straight-empty.png",
            'curve': "../assets/pipes-event/pipe-curve.png",
            'curve-empty': "../assets/pipes-event/pipe-curve-empty.png"
        }

        # Center the game
        gameOffset = (parentDiv.size - gameSz) * 0.5

        self.__tiles = [[None for y in range(self.__numTiles[1])] for x in range(self.__numTiles[0])]
        endpos = (self.__numTiles[0] - 1, 0)
        pos = (0, self.__numTiles[1] - 1)
        origin = (-1, self.__numTiles[1] - 1)
        lastpath = []
        anchors = [] # points where we can go more than 1 way
        origins = []
        paths = [] # paths between anchors
        while True:
            tilePos = gameOffset + avg.Point2D(pos[0] * tileSz[0], pos[1] * tileSz[1])
            angle = random.choice([0, 90, 180, 270])
            tile = self.__tiles[pos[0]][pos[1]] = avg.ImageNode(href=self.__assets['straight'], pos=tilePos, size=tileSz, angle=math.radians(angle), parent=parentDiv)

            tile.correctAngles = []
            tile.tileType = 'straight'
            tile.tileOrigin = origin

            if pos == endpos:
                newpos = (self.__numTiles[0], 0)
            else:
                dirs = [(1, 0), (-1, 0), (0, 1), (0, -1)]
                dirs = [d for d in dirs if self.__isPossibleDir(pos, d)]

                if not dirs:
                    if not anchors:
                        print "??? No more anchors to track back to @" + `pos` + " ???"
                        break

                    for p in lastpath:
                        self.__tiles[p[0]][p[1]].unlink()
                        self.__tiles[p[0]][p[1]] = None

                    pos = anchors.pop()
                    origin = origins.pop()
                    lastpath = paths.pop()
                    continue

                rdir = random.choice(dirs)
                newpos = (pos[0] + rdir[0], pos[1] + rdir[1])

            neworigin = pos

            correctAngles = [0]
            assetDir = (newpos[0] - origin[0], newpos[1] - origin[1])
            if assetDir[1] == 0:
                asset = 'straight'
                correctAngles = [0, 180]
            elif assetDir[0] == 0:
                asset = 'straight'
                correctAngles = [90, 270]
            else:
                asset = 'curve'
                olddir = (pos[0] - origin[0], pos[1] - origin[1])
                rotated = 0
                testdir1 = avg.Point2D(0, 1)
                testdir2 = avg.Point2D(1, 0)
                p = avg.Point2D(pos[0], pos[1])
                while not ((p + testdir1 == newpos and p + testdir2 == origin) or (p + testdir2 == newpos and p + testdir1 == origin)):
                    testdir1 = self.__rotateVec90CW(testdir1)
                    testdir2 = self.__rotateVec90CW(testdir2)
                    rotated += 90

                correctAngles = [rotated]

            tile.correctAngles = correctAngles
            tile.href = self.__assets[asset]
            tile.tileType = asset

            if pos == endpos:
                break

            pos = newpos
            origin = neworigin

            if len(dirs) > 1:
                anchors.append(pos)
                origins.append(origin)
                paths.append(lastpath)
                lastpath = []

            lastpath.append(pos)

        # Fill non-filled tiles with random stuff and subscribe click handlers
        for x in range(self.__numTiles[0]):
            for y in range(self.__numTiles[1]):
                if self.__tiles[x][y] == None:
                    asset = random.choice(['curve', 'straight'])
                    angle = random.choice([0, 90, 180, 270])
                    tilePos = gameOffset + avg.Point2D(x * tileSz[0], y * tileSz[1])
                    tile = self.__tiles[x][y] = avg.ImageNode(href=self.__assets[asset], pos=tilePos, size=tileSz, angle=math.radians(angle), parent=parentDiv)
                    tile.correctAngles = []
                    tile.tileType = asset

                self.__tiles[x][y].subscribe(avg.Node.CURSOR_DOWN, lambda event,xx=x,yy=y: self.__onClickTile(xx, yy))

        # Add start and end pipes
        self.__startPipe = avg.ImageNode(href=self.__assets['straight'], pos=gameOffset + avg.Point2D(-tileSz[0], (self.__numTiles[1] - 1) * tileSz[1]), size=tileSz, parent=parentDiv)
        self.__endPipe = avg.ImageNode(href=self.__assets['straight-empty'], pos=gameOffset + avg.Point2D(self.__numTiles[0] * tileSz[0], 0), size=tileSz, parent=parentDiv)

        avgPlayer.subscribe(avgPlayer.KEY_UP, self.__onKeyUp)

        self.__isComplete()

    def __rotateVec90CW(self, vec):
        v = (vec.x, vec.y)
        r = {
            (1, 0): (0, 1),
            (0, 1): (-1, 0),
            (-1, 0): (0, -1),
            (0, -1): (1, 0)
        }[v]
        return avg.Point2D(r[0], r[1])

    def __isPossibleDir(self, atPos, dir):
        newpos = (atPos[0] + dir[0], atPos[1] + dir[1])
        if self.__isOutOfBounds(newpos[0], newpos[1]):
            return False

        if self.__tiles[newpos[0]][newpos[1]]:
            return False

        return True

    def __isOutOfBounds(self, x, y):
        return x < 0 or x >= self.__numTiles[0] or y < 0 or y >= self.__numTiles[1]

    def __isComplete(self):
        # Make all tiles empty
        # TODO: We could avoid changing href here alot!
        for x in range(self.__numTiles[0]):
            for y in range(self.__numTiles[1]):
                tile = self.__tiles[x][y]
                tile.href = self.__assets[tile.tileType + "-empty"]

        endpos = (self.__numTiles[0], 0)
        pos = (0, self.__numTiles[1] - 1)
        origdir = (1, 0)
        anchors = [] # curve tiles where it was unclear in which direction to go
        anchorOrigDirs = [] # origdirs for each anchor
        wrapBack = False
        numwb = 0
        while pos != endpos:
            if self.__isOutOfBounds(pos[0], pos[1]):
                return False

            tile = self.__tiles[pos[0]][pos[1]]
            angle = int(math.floor(math.degrees(tile.angle)))
            if tile.tileType == 'straight':
                if angle % 180 == 0:
                    if not origdir[1] == 0:
                        return False
                elif angle % 90 == 0:
                    if not origdir[0] == 0:
                        return False

                tile.href = self.__assets[tile.tileType]
                pos = (pos[0] + origdir[0], pos[1] + origdir[1])
            else:
                # inpointing directions based on the curve asset
                t1 = avg.Point2D(-1,0)
                t2 = avg.Point2D(0,-1)
                while angle > 0:
                    t1 = self.__rotateVec90CW(t1)
                    t2 = self.__rotateVec90CW(t2)
                    angle -= 90

                t1 = (int(t1.x), int(t1.y))
                t2 = (int(t2.x), int(t2.y))
                if t1 == origdir:
                    t2 = (-t2[0], -t2[1])
                    origdir = t2
                    tile.href = self.__assets[tile.tileType]
                    pos = (pos[0] + t2[0], pos[1] + t2[1])
                elif t2 == origdir:
                    t1 = (-t1[0], -t1[1])
                    origdir = t1
                    tile.href = self.__assets[tile.tileType]
                    pos = (pos[0] + t1[0], pos[1] + t1[1])
                else:
                    return False

        endtile = self.__tiles[endpos[0] - 1][endpos[1]]
        endtile.href = self.__assets[endtile.tileType]
        self.__endPipe.href = self.__assets['straight']
        return True

    def __onKeyUp(self, event):
        if (event.keyname == 'S'):
            self.__solve()

    def __solve(self):
        for x in range(self.__numTiles[0]):
            for y in range(self.__numTiles[1]):
                tile = self.__tiles[x][y]
                if tile.correctAngles:
                    tile.angle = math.radians(tile.correctAngles[0])

        self.__isComplete()

    def __onClickTile(self, x, y):
        self.__tiles[x][y].angle += math.radians(90)
        if self.__isComplete():
            for tiles in self.__tiles:
                for tile in tiles:
                    tile.sensitive = False
            avg.Player.get().setTimeout(2000, self.__onCompleted)

    def onFrame(self, dt, keystates):
        if self.__completed:
            return True

    def __onCompleted(self):
        self.__completed = True
