from libavg import avg, player as avgPlayer
from item.Brakefluid import Brakefluid
from item.Inventory import Inventory

class MovementControls:

    SPEED_INIT = 0.0 # Initial start speed [pixel/s]
    SPEED_MIN = 50.0 # minimum speed when using breaks [pixel/s]
    SPEED_MAX = 400.0 # maximum movement speed [pixels/s]
    DISABLE_THRESHOLD = 8000 # how long breaks can be used at max intensity before breaking [ms]
    DISABLED_TIME = 3000 # how long breaks will be disabled when they are broken [ms]
    RESET_TIME = 500 # how long it takes for lever to reset to minimum intensity when there is no contact [ms]
    ACCELERATION = 250 # [pixel/s^2]
    BRAKE_DEFAULT_ACC_RATIO = 2.4 # Ratio of maximum break acceleration to forward acceleration
    BRAKE_FLUID_ACC_RATIO = 4.0 # Ratio of maximum break acceleration when using Brakefluid to forward acceleration

    def __init__(self, parentDiv, player, inventory):
        self.__player = player

        self.active = True
        self.reset()

        bhbw = 15 # breakHealthBarWidth
        lh = 50 # leverHeight
        relOff = 0.9 # Offset of leverMotionDiv to leverDiv relative to lever size

        self.__mainDiv = avg.DivNode(pos=(25,0), size=(125 + bhbw,270), parent=parentDiv)
        self.__leverDiv = avg.DivNode(pos=(0,25), size=self.__mainDiv.size - (bhbw,25), parent=self.__mainDiv)
        self.__bgTexture = avg.ImageNode(href="../assets/HUD/brake_background.png", size=self.__leverDiv.size, parent=self.__leverDiv)
        self.__leverMotionDiv = avg.DivNode(pos=(0,lh * relOff), size=self.__leverDiv.size - avg.Point2D(0, lh * relOff * 2), parent=self.__leverDiv) # valid dragging area
        self.__lever = avg.ImageNode(href="../assets/HUD/brake_handle.png", pos=(0,self.__leverMotionDiv.size.y - lh), size=(self.__leverDiv.size.x,lh), parent=self.__leverDiv)
        self.__debugString = avg.WordsNode(pos=(0,self.__mainDiv.height), sensitive=False, active=False, parent=self.__mainDiv)

        self.__breakHealthGlow = avg.ImageNode(size=player.imageDiv.size, href="../assets/player-glow.png", opacity=0, parent=player.imageDiv)
        self.__breakHealthBar = avg.RectNode(pos=(self.__leverDiv.width,0), size=(bhbw,0), opacity=0, fillopacity=1, fillcolor="FF0000", sensitive=False, parent=self.__leverDiv)

        # calculate only once
        self.__leverMotionDivGlobalOffsetY = self.__leverMotionDiv.pos.y + self.__leverDiv.pos.y
        self.__leverOffsetY = (relOff - 0.5) * self.__lever.size.y

        if avgPlayer.drawHelpers:
            self.__debugString.active = True
            self.__mainDiv.elementoutlinecolor = "FFFF00"

        self.__leverContact = None
        self.__lever.subscribe(avg.Node.CURSOR_DOWN, self.__onLeverDown)
        self.__lever.subscribe(avg.Node.CURSOR_MOTION, self.__onLeverMotion)
        self.__mainDiv.subscribe(avg.Node.CURSOR_MOTION, self.__onLeverMotion)
        self.__mainDiv.subscribe(avg.Node.CURSOR_UP, self.__onLeverUp)
        self.__mainDiv.subscribe(avg.Node.CURSOR_OUT, self.__onLeverOut)
        self.__mainDiv.subscribe(avg.Node.CURSOR_OVER, self.__onLeverIn)

        self.__inventory = inventory
        self.__inventory.subscribe(Inventory.ITEM_ACTIVATED, self.__onItemActivate)
        self.__inventory.subscribe(Inventory.ITEM_DEACTIVATED, self.__onItemDeactivate)

    def reset(self):
        self.__player.speed = MovementControls.SPEED_INIT # current speed [pixel/s]
        self.__breakIntensity = 0.0 # how much the break is used / valid range: 0 to 1
        self.__breakActiveTime = 0 # how long the break has been used [ms]
        self.__breakDisabledTime = 0 # how long the break is disabled [ms]
        self.__breakDisabled = False
        self.__frameCounter = 0
        self.__breakHealth = 0.0 # ratio breakActiveTime to DISABLE_THRESHOLD / valid range: 0 to 1
        self.__brakeAccRatio = MovementControls.BRAKE_DEFAULT_ACC_RATIO # Ratio of maximum break acceleration to forward acceleration

    def destroy(self):
        self.__mainDiv.unlink(True)
        self.__mainDiv = None
        self.__breakHealthGlow.unlink(True)
        self.__breakHealthGlow = None
        self.__player = None
        self.__inventory = None

    def __onLeverDown(self, event):
        self.__leverContact = event.contact

        # allow only movement and touch events while in valid area
        posY = event.pos.y - self.__leverMotionDivGlobalOffsetY
        if not self.__breakDisabled:
            if posY > self.__leverMotionDiv.size.y:
                self.__breakIntensity = 0.0
            elif posY < 0:
                self.__breakIntensity = 1.0
            else:
                self.__breakIntensity = 1 - posY / self.__leverMotionDiv.size.y

    # Mouse interaction
    def __onLeverMotion(self, event):
        if self.__leverContact:
            self.__onLeverDown(event)

    def __onLeverUp(self, event):
        self.__leverContact = None

    def __onLeverOut(self, event):
        # Check if we leave the lever downwards
        if (self.__mainDiv.pos.x + self.__leverDiv.width > event.pos.x > self.__mainDiv.pos.x and event.pos.y > self.__mainDiv.pos.y + self.__mainDiv.height):
            self.__onLeverMotion(event)
        else:
            # Release contact if not
            self.__onLeverUp(event)

    def __onLeverIn(self, event):
        # If we enter the leverDiv on the same side (bottom) we left it
        if self.__mainDiv.pos.x + self.__leverDiv.width > event.pos.x > self.__mainDiv.pos.x and event.pos.y > self.__mainDiv.pos.y + self.__mainDiv.height - 20:
            # Continue evaluating userinput
            self.__onLeverMotion(event)
        else:
            # Release contact if not
            self.__onLeverUp(event)

    def show(self, show=True):
        if not self.active and show:
            return
        if self.__mainDiv:
            self.__mainDiv.active = show

    def onFrame(self, dt):
        if self.__frameCounter < 2:
            self.__frameCounter += 1
            return
        acc = dt * 0.001 * MovementControls.ACCELERATION

        # reenable breaks after time
        if self.__breakDisabled:
            self.__breakDisabledTime += dt

            self.__breakHealth = 1.0 - self.__breakDisabledTime / MovementControls.DISABLED_TIME

            if self.__breakDisabledTime >= MovementControls.DISABLED_TIME:
                self.__breakDisabled = False
                self.__breakDisabledTime = 0
                self.__breakActiveTime = 0

        else:
            if self.__breakIntensity > 0:
                # add break usage time relative to break intensity
                self.__breakActiveTime += dt * self.__breakIntensity

                if not self.__leverContact:
                    # Automatically lower break intensity when no user interaction is present
                    self.__breakIntensity = max(self.__breakIntensity - dt / MovementControls.RESET_TIME, 0.0)
            else:
                # cool down twice as fast as overheat
                self.__breakActiveTime = max(self.__breakActiveTime - 2 * dt, 0)

            self.__breakHealth = self.__breakActiveTime / MovementControls.DISABLE_THRESHOLD

            # detect overheating
            if self.__breakActiveTime >= MovementControls.DISABLE_THRESHOLD:
                self.__breakDisabled = True
                self.__breakIntensity = 0.0

        # Accelerate the player automatically
        self.__player.speed += acc

        # Apply breaks
        tmpSpeed = self.__player.speed - self.__breakIntensity * acc * self.__brakeAccRatio
        self.__player.speed = (tmpSpeed if tmpSpeed > MovementControls.SPEED_MIN else MovementControls.SPEED_MIN)

        # limit to maximum / minimum speed
        self.__player.speed = min(self.__player.speed, MovementControls.SPEED_MAX)

        # update visuals
        self.__lever.pos = avg.Point2D(0, (1 - self.__breakIntensity) * self.__leverMotionDiv.size.y + self.__leverOffsetY)
        self.__breakHealthGlow.opacity = self.__breakHealth
        self.__breakHealthBar.size = (self.__breakHealthBar.size.x, self.__leverDiv.height * self.__breakHealth)

        # debug
        self.__debugString.text = str(round(self.__player.speed, 1)) + " | " + str(round(self.__breakIntensity, 1)) + " | " + str(round(self.__breakActiveTime, 1)) + " | " + str(round(self.__breakDisabledTime, 0)) + " | " + str(round(self.__breakHealth, 1)) + " | " + str(round(self.__brakeAccRatio, 1))

        # Update player speed
        self.__player.speed = self.__player.speed

    def getSpeed(self):
        return self.__player.speed

    def __onItemActivate(self, item):
        if item.__class__ == Brakefluid:
            self.__brakeAccRatio = MovementControls.BRAKE_FLUID_ACC_RATIO

    def __onItemDeactivate(self, item):
        if item.__class__ == Brakefluid:
            self.__brakeAccRatio = MovementControls.BRAKE_DEFAULT_ACC_RATIO
