from libavg import avg, logger, player as avgPlayer
import Util
from Item import *
from Petroleum import *
from StatisticsManager import StatisticsManager

# The current inventory of the player
# Collects all items, activated and not activated
# Displays collected items on the HUD
class Inventory(avg.Publisher):
    HEIGHT = 150 # [px]
    COLOR_INACTIVE = "fed976" # Background color for inactive items [color]
    COLOR_ACTIVE = "31a354" # Background color for active items [color]
    ITEM_SIZE = (120,120) # quadratic size of nodes [(px,px)]
    ITEM_OFFSET_Y = (HEIGHT - ITEM_SIZE[1]) * 0.5
    ITEM_BORDER = 10 # space between nodes [px]
    MAX_ITEMS = 5 # maximum number of items in inventory
    NEW_ITEM_COLLECTED = avg.Publisher.genMessageID() # An item, that has been collected for the first time
    ITEM_ACTIVATED = avg.Publisher.genMessageID() # An Item has been activated
    ITEM_DEACTIVATED = avg.Publisher.genMessageID() # An Item has been deactivated

    def __init__(self, parentDiv, offsetX):
        # Init Message System
        super(Inventory, self).__init__()
        self.publish(Inventory.NEW_ITEM_COLLECTED)
        self.publish(Inventory.ITEM_ACTIVATED)
        self.publish(Inventory.ITEM_DEACTIVATED)

        self.__items = {} # Dictonary accumulating all present item instances and their related avg Nodes (avg.DivNode: item)
        self.__activeItems = {} # The classes of all currently active items and their running instance

        # Visual representation
        self.__parentDiv = parentDiv
        width = Inventory.MAX_ITEMS * (Inventory.ITEM_SIZE[0] + Inventory.ITEM_BORDER) + Inventory.ITEM_BORDER
        self.__mainDiv = avg.DivNode(pos=(offsetX - width,parentDiv.height - Inventory.HEIGHT - 25), size=(width,Inventory.HEIGHT), parent=parentDiv)

    def destroy(self):
        while len(self.__items) > 0:
            self.__items.popitem()[0].unlink(True)
        self.__items = None
        self.__mainDiv.unlink(True)
        self.__mainDiv = None

    def show(self, show):
        if self.__mainDiv:
            self.__mainDiv.active = show

    # Add item to inventory
    # Returns True on success, False otherwise
    def addItem(self, item):
        if not isinstance(item, Item):
            logger.error("Cannot add Object " + str(item) + ". Not an Item!")
            return False
        if len(self.__items) >= Inventory.MAX_ITEMS:
            logger.info("Cannot add item: Maximum number of " + str(Inventory.MAX_ITEMS) + " items reached.")
            return False

        divNode = avg.DivNode(size=Inventory.ITEM_SIZE, parent=self.__mainDiv)
        itemHealthBar = avg.RectNode(size=divNode.size, fillcolor=Inventory.COLOR_INACTIVE, color=Inventory.COLOR_INACTIVE, fillopacity=0.2, opacity=0.7, parent=divNode)
        imgNode = avg.ImageNode(href=item.getImageHref(), size=divNode.size, parent=divNode)

        divNode.index = len(self.__items) # First free index by design
        divNode.pos = (self.__getItemOffsetX(divNode.index), Inventory.ITEM_OFFSET_Y)
        divNode.subscribe(avg.Node.CURSOR_DOWN, lambda e: self.__onItemCursorDown(divNode, e))

        if not item.itemTypeCollected():
            self.notifySubscribers(Inventory.NEW_ITEM_COLLECTED, [item.__class__])
            item.collectItem()

        self.__items[divNode] = item
        avgPlayer.statisticsManager.addValue(StatisticsManager.C_ITEMS_COLLECTED, item.getItemName(), 1)

        return True

    # Returns the active item of an Item Class if there is any, else None
    def getActiveItemOfClass(self, itemClass):
        if itemClass in self.__activeItems:
            return self.__activeItems[itemClass]
        else:
            return None

    def onFrame(self, dt):
        itemsToRemove = [] # Delete item after dict traversion to avoid dict size change during iteration
        for node, item in self.__items.iteritems():

            item.onFrame(dt)

            # Update indicators
            if item.isActive():
                rectNode = Util.getFirstChildNodeofClass(node, avg.RectNode)
                height = (1 - item.getActiveTimeRatio()) * rectNode.parent.height
                rectNode.pos = (rectNode.pos.x, rectNode.parent.height - height)
                rectNode.size = (rectNode.size.x, height)

            # Check for expired items and remove them
            if item.isExpired():
                itemsToRemove.append(item)

        if itemsToRemove:
            # Allow multiple items to expire on the same frame
            for itemToRemove in itemsToRemove:
                del self.__activeItems[itemToRemove.__class__]
                self.__removeItem(itemToRemove)
        del itemsToRemove[:]

    # Invoked by pressing an items icon
    def __onItemCursorDown(self, divNode, event):
        item = self.__items[divNode]
        rectNode = Util.getFirstChildNodeofClass(divNode, avg.RectNode)
        itemClass = item.__class__
        if not item.isActive():
            if not itemClass in self.__activeItems:
                self.__activeItems[itemClass] = item
                rectNode.color = Inventory.COLOR_ACTIVE
                rectNode.fillcolor = Inventory.COLOR_ACTIVE
                item.activate()
                self.notifySubscribers(Inventory.ITEM_ACTIVATED, [item])
        else:
            del self.__activeItems[itemClass]
            rectNode.color = Inventory.COLOR_INACTIVE
            rectNode.fillcolor = Inventory.COLOR_INACTIVE
            item.deactivate()
            self.notifySubscribers(Inventory.ITEM_DEACTIVATED, [item])

    # Calculate pixel offset of an item on x axis for given index
    # Items are placed right to left
    def __getItemOffsetX(self, index):
        return self.__mainDiv.width - ((index+1) * (Inventory.ITEM_SIZE[0] + Inventory.ITEM_BORDER))

    # Remove item and its visual representation
    def __removeItem(self, item):
        node = None
        for nodeKey, itemValue in self.__items.iteritems():
            if itemValue == item:
                node = nodeKey
        deletedIndex = node.index
        node.unlink(True)
        del self.__items[node]
        node = None

        # Reorder following items
        for node in self.__items:
            if node.index > deletedIndex:
                node.index -= 1
                node.pos = (self.__getItemOffsetX(node.index), Inventory.ITEM_OFFSET_Y)

        avgPlayer.statisticsManager.addValue(StatisticsManager.C_ITEMS_FULLY_USED, item.getItemName(), 1)
        self.notifySubscribers(Inventory.ITEM_DEACTIVATED, [item])
