import math
from abc import ABCMeta, abstractmethod
from libavg import avg, logger
from Tile import TileFeature

# ItemFeatures are spwaned as TileFeatures and can be collected once when the player is passing by
# This is the representation of items on the level without any player interaction
class ItemTileFeature(TileFeature):
    __metaclass__ = ABCMeta
    FEATURE_NAME = "default_itemTileFeature"

    def __init__(self, tile):
        super(ItemTileFeature, self).__init__(tile)
        href = self.getImageHref()
        self.__image = avg.ImageNode(href=href, pos=tile.pos, size=tile.size, angle = math.radians(tile.rotation), sensitive=False, parent=tile.parent)

    # Overwrite this method in your itemTileFeature implementation
    @abstractmethod
    def getFeatureName(self):
        return Item.FEATURE_NAME

    # Overwrite this method to return desired item object
    @abstractmethod
    def getItemInstance(self):
        return Item()

    # Overwrite this method in your itemTileFeature implementation if needed
    # Visual representation on world
    def getImageHref(self):
        return "../assets/items/chest_item.png"

    # Called once everytime the player enters the tile of this ItemTileFeature
    # This item is either added to the inventory and the TileFeature is removed, or, if not possible, nothing happens
    def onPlayerEnteredTile(self, **kwargs):
        if not "inventory" in kwargs:
            logger.error("No inventory instance provided")
            return
        inventory = kwargs["inventory"]
        success = inventory.addItem(self.getItemInstance())
        if success:
            self.removeItemFeature()

    def removeItemFeature(self):
        self.getTile().removeFeature(self.getFeatureName())

    def onDestroy(self):
        self.__image.unlink(True)
        self.__image = None



# Defines an item itself
# All items are collected in the players inventory as long as they are not expired
class Item():
    __metaclass__ = ABCMeta
    ITEM_NAME = "default_item"
    DURATION = 10000 # How long one instance of this item can be active before it expires [ms]
    COLLECTED_ONCE = False # If an item of this class has been collected once


    def __init__(self):
        self.__active = False # If this item is currently in use
        self.__activeTime = 0 # How long this item has been used allready [ms]
        self.__expired = False # If this item can be used

    # Overwrite this method in your Item implementation
    @abstractmethod
    def getItemName(self):
        return ITEM_NAME

    # Overwrite this method in your Item implementation
    # Visual representation in the inventory
    @abstractmethod
    def getImageHref():
        return "../assets/mine/event.png"

    # Use the effect of this item
    def activate(self):
        self.__active = True

    # Pause using this item
    def deactivate(self):
        self.__active = False

    def isActive(self):
        return self.__active

    def expire(self):
        self.__expired = True

    def isExpired(self):
        return self.__expired

    # Time this item has been active [ms]
    def getActiveTime(self):
        return self.__activeTime

    # Ratio of allready used time to overall time this item can be active
    def getActiveTimeRatio(self):
        return self.__activeTime / self.__class__.DURATION

    # Increase usage time if item is used
    # Mark as expired if item is used up
    def onFrame(self, dt):
        if self.__active:
            self.__activeTime += dt
            if self.__activeTime > self.__class__.DURATION:
                self.__expired = True

    # Returns if an item of this class has been collected
    def itemTypeCollected(self):
        return self.__class__.COLLECTED_ONCE

    def collectItem(self):
        self.__class__.COLLECTED_ONCE = True
