from libavg import avg
from menu.StyleDefinitions import StyleDefinitions
import time

class Button(avg.DivNode):

    BUTTON_DOWN = avg.Publisher.genMessageID()
    BUTTON_UP = avg.Publisher.genMessageID()
    BUTTON_CLICKED = avg.Publisher.genMessageID()
    BUTTON_CLICKED_DELAY = avg.Publisher.genMessageID()

    def __init__(self, parent, text="BUTTON", toggleText=None, font=StyleDefinitions.FONT_HEADLINE, size=StyleDefinitions.BUTTON_LARGE_SIZE, pos=avg.Point2D(0,0), textColor=StyleDefinitions.BACKGROUND_COLOR, backgroundColor=StyleDefinitions.FONT_COLOR, opacity=1.0, bgImageHref=StyleDefinitions.BUTTON_BG_IMAGE, delay=None):
        super(Button, self).__init__(pos=pos, size=size)
        self.registerInstance(self, parent)

        self.publish(Button.BUTTON_DOWN)
        self.publish(Button.BUTTON_UP)
        self.publish(Button.BUTTON_CLICKED)
        self.publish(Button.BUTTON_CLICKED_DELAY)

        self.__shadow = avg.ShadowFXNode(radius=20, color="000000", opacity=0.6)
        self.__mainDiv = avg.DivNode(size=self.size, parent=self)
        self.__buttonBg = avg.RectNode(size=self.__mainDiv.size, color=backgroundColor, fillcolor=backgroundColor, fillopacity=opacity, opacity=opacity, parent=self.__mainDiv)
        self.__bgImage = avg.ImageNode(size=self.__mainDiv.size, parent=self.__mainDiv)
        self.__buttonProgress = avg.RectNode(opacity=0.9, color=StyleDefinitions.HIGHLIGHT_COLOR, fillopacity=0.7, fillcolor=StyleDefinitions.HIGHLIGHT_COLOR, parent=self.__mainDiv)
        self.__buttonText = avg.WordsNode(font=font, text=text, alignment="center", color=textColor, parent=self.__mainDiv)

        self.setBgImage(bgImageHref)

        self.setFont(font)

        self.setMaxTextSize()

        self.__delay = delay # How long it takes to notify subscribers about an BUTTON_CLICKED_DELAY event [ms]
        self.__downTime = 0 # Time this button was pressed [SystemTime]
        self.__currentlyPressed = False
        self.subscribe(avg.Node.CURSOR_DOWN, self.__onButtonDown)
        self.subscribe(avg.Node.CURSOR_UP, self.__onButtonUp)
        self.subscribe(avg.Node.CURSOR_OUT, self.__onCursorOut)

        self.__touchID = None

        self.__getTime = lambda: int(round(time.time() * 1000))

    # Set new Text to Button
    # Text size will automatically be adjusted
    def setText(self, text):
        self.__buttonText.text = text
        self.setMaxTextSize()

    # Centeres text inside the button
    def centerText(self):
        extents = self.__buttonText.getLineExtents(0)
        self.__buttonText.pos = (self.width * 0.5, self.height * 0.5 - extents.y * 0.5)

    # Propagate Size property to every child element
    def setSize(self, size):
        self.size = size
        self.__mainDiv.size = size
        self.__buttonBg.size = size
        self.__bgImage.size = size
        self.setMaxTextSize()

    # Sets fontSize accordingly to text length and button length
    def setMaxTextSize(self):
        fontSize = StyleDefinitions.BUTTON_LARGE_FONTSIZE
        if self.size != StyleDefinitions.BUTTON_LARGE_SIZE:
            # Set fontsize to buttonwidth ratio equally to all sizes
            ratio = self.size.x / StyleDefinitions.BUTTON_LARGE_SIZE.x
            fontSize *= ratio
        self.__buttonText.fontsize = fontSize

        # If text is too long, decrese fontsize until its width is smaller than surrounding button plus additional margin
        while self.__buttonText.size.x > self.width - 30:
            self.__buttonText.fontsize -= 0.5

        self.centerText()

    def getTextSize(self):
        return self.__buttonText.fontsize

    def setTextSize(self, fontSize):
        self.__buttonText.fontsize = fontSize

    def setBgImage(self, bgImageHref):
        if isinstance(bgImageHref, unicode) or isinstance(bgImageHref, str):
            self.__buttonBg.active = False
            self.__bgImage.href = bgImageHref
            self.__bgImage.setEffect(self.__shadow)
            self.__hasBgImage = True
        else:
            self.__hasBgImage = False

    def setFont(self, fontName):
        if isinstance(fontName, unicode) or isinstance(fontName, str):
            self.__buttonText.font = fontName

    def __onButtonDown(self, e):
        self.notifySubscribers(Button.BUTTON_DOWN, [])
        if self.__touchID == None:
            self.__downTime = self.__getTime()
            self.__touchID = e.cursorid
            if self.__delay:
                self.__progressAnim = avg.LinearAnim(node=self.__buttonProgress, attrName="size", duration=self.__delay, startValue=(0,self.height), endValue=(self.width, self.height))
                self.__progressAnim.start()
            self.__toggleVisuals()

    def __onButtonUp(self, e):
        self.notifySubscribers(Button.BUTTON_UP, [])
        if e.cursorid == self.__touchID:
            self.notifySubscribers(Button.BUTTON_CLICKED, [])
            if self.__delay:
                self.__resetProgress()
                if self.__getTime() - self.__downTime > self.__delay:
                    self.notifySubscribers(Button.BUTTON_CLICKED_DELAY, [])
            self.__toggleVisuals()
            self.__touchID = None

    def __onCursorOut(self, e):
        if e.cursorid == self.__touchID:
            self.__toggleVisuals()
            self.__touchID = None
            if self.__delay:
                self.__resetProgress()

    def __toggleVisuals(self):
        if self.__hasBgImage:
            # Add press effect by moving the button minimally
            self.__mainDiv.pos = (0,0) if self.__mainDiv.pos != avg.Point2D(0,0) else (5,5)
        else:
            # Swap background and text color
            self.__buttonBg.color, self.__buttonBg.fillcolor, self.__buttonText.color = self.__buttonText.color, self.__buttonText.color, self.__buttonBg.color

    def __resetProgress(self):
        self.__progressAnim.abort()
        self.__buttonProgress.size = (0,0)
