####################################################################################
#
# Tilemap Editor for MazeRunner Game
#
####################################################################################

from EdConstants import *

from abc import ABCMeta, abstractmethod
from EdWorld import *
from World import *
from event.SpearTrap import *
from event.Portal import *
from Util import *

####################################################################################

class Tool():
    __metaclass__ = ABCMeta

    def __init__(self, world):
        self.world = world
        self.__enabled = False

    # Returns the identifying name of the tool implementation
    @abstractmethod
    def getName(self):
        return "unknown_tool"

    # Returns the description of the tool implementation
    @abstractmethod
    def getDescription(self):
        return "Unknown Tool"

    def onCursorMotion(self, event):
        pass

    def onWorldCursorMotion(self, event):
        pass

    def onTileClicked(self, event, tile):
        pass

    # User selects this tool
    def onEnable(self):
        pass

    # User selects another tool
    def onDisable(self):
        pass

    def isEnabled(self):
        return self.__enabled

    def enable(self, enable=True):
        self.__enabled = enable
        if enable:
            self.onEnable()
        else:
            self.onDisable()

    def disable(self):
        self.enable(False)

    enabled = property(isEnabled, enable)

####################################################################################

class PathTool(Tool):
    NAME = "path_tool"

    # brush - True to enable brushing, False to place single tiles
    def __init__(self, world, brush=False):
        super(PathTool, self).__init__(world)
        self.__brush = brush

    def getName(self):
        return PathTool.NAME

    def getDescription(self):
        return "Place/Remove path tile"

    def __placePathTile(self, tile):
        tile.texture = 'path_straight'
        self.world.invalidateSurroundings(self.world.getTilePos(tile))

    def __removePathTile(self, tile):
        tile.texture = 'ground'
        self.world.invalidateSurroundings(self.world.getTilePos(tile))

    def onWorldCursorMotion(self, event):
        if self.__brush and (event.leftbuttonstate or event.rightbuttonstate):
            tile = self.world.getTile(event.tilePos)
            if tile:
                if event.leftbuttonstate:
                    self.__placePathTile(tile)
                elif event.rightbuttonstate:
                    self.__removePathTile(tile)

    def onTileClicked(self, event, tile):
        if event.lmouse:
            self.__placePathTile(tile)
        elif event.rmouse:
            self.__removePathTile(tile)
        else:
            return

####################################################################################

class PathBrushTool(PathTool):
    NAME = "path_brush_tool"

    def __init__(self, world):
        super(PathBrushTool, self).__init__(world, True)

    def getName(self):
        return PathBrushTool.NAME

    def getDescription(self):
        return "Path brush"

####################################################################################

class PlaceTool(Tool):
    NAME = "place_tool"

    def __init__(self, world, menuDiv):
        super(PlaceTool, self).__init__(world)
        self.__objectName = ""
        self.__menu = menuDiv
        self.__objectButtons = {}
        self.__prevFeature = None # Set after placing the first event feature

    # If isprototype is True, when placing the object, the given object will
    # be used as a prototype for the tilevent that will be added as a TileFeature
    # If isprototype is False, the exact given object can be placed.
    def registerObject(self, name, description, iseventprototype=False, hidden=False):
        btnSz = (90,22)
        btnPadding = 5

        numVisibleButtons = 0
        for btnName in self.__objectButtons:
            if not self.__objectButtons[btnName].hidden:
                numVisibleButtons += 1

        pos = avg.Point2D(numVisibleButtons % 2, numVisibleButtons // 2) * avg.Point2D(btnSz[0] + btnPadding, btnSz[1] + btnPadding)
        btn = avg.DivNode(pos=pos, active=False, parent=self.__menu)
        btn.objectName = name
        btn.isEventPrototype = iseventprototype
        btn.hidden = hidden

        if not hidden:
            btn.btnBG = avg.RectNode(size=btnSz, strokewidth=2, color="FFFFFF", fillcolor="444444", fillopacity=1.0, parent=btn)
            btn.btnText = avg.WordsNode(pos=(btnSz[0]*0.5,4), text=description, fontsize=13, color="FFFFFF", alignment="center", parent=btn)

            btn.subscribe(avg.Node.CURSOR_DOWN, lambda e,btn=btn: self.__onObjectButtonClicked(btn))
            btn.subscribe(avg.Node.CURSOR_OVER, lambda e,btnBG=btn.btnBG: self.__onObjectButtonBackgroundHoverStateChanged(btnBG, True))
            btn.subscribe(avg.Node.CURSOR_OUT, lambda e,btnBG=btn.btnBG: self.__onObjectButtonBackgroundHoverStateChanged(btnBG, False))

        self.__objectButtons[name] = btn

    def getName(self):
        return PlaceTool.NAME

    def getDescription(self):
        return "Place objects"

    def setPlaceable(self, name):
        if self.__objectName:
            obj = self.world.getObject(self.__objectName)
            if obj:
                obj.hideHelper()

        self.__objectName = name
        obj = self.world.getObject(self.__objectName)
        if obj and self.isEnabled():
            obj.showHelper()

    # Returns the name of the placeable that is currently selected
    def getPlaceable(self):
        return self.__objectName

    def onEnable(self):
        obj = self.world.getObject(self.__objectName)
        if obj:
            obj.showHelper()

        for btnName in self.__objectButtons:
            btn = self.__objectButtons[btnName]
            btn.active = not btn.hidden

    def onDisable(self):
        obj = self.world.getObject(self.__objectName)
        if obj:
            obj.hideHelper()

        for btnName in self.__objectButtons:
            self.__objectButtons[btnName].active = False

    def onCursorMotion(self, event):
        obj = self.world.getObject(self.__objectName)
        if obj:
            obj.setHelperPos(event.tilePos)

    # Returns True if an object was actually placed, False if not
    def onTileClicked(self, event, tile):
        if tile.texture != 'ground':
            if event.lmouse:
                obj = self.world.getObject(self.__objectName)
                if obj:
                    btn = self.__objectButtons[self.__objectName]
                    if btn and btn.isEventPrototype:
                        feature = tile.getFeature(EventFeature.FEATURE_NAME)
                        if feature:
                            feature.destroy()
                            tile.removeFeature(EventFeature.FEATURE_NAME)

                        feature = EventFeature(tile, self.__objectName, obj)
                        tile.addFeature(feature)

                        isParent = False
                        if self.__prevFeature is not None:
                            isParent = (self.__objectName == SpearTrapLever.FEATURE_NAME)
                            isParent = isParent or (self.__objectName == Portal.FEATURE_NAME and self.__prevFeature.getEventName() == Portal.FEATURE_NAME and not self.__prevFeature.hasParent())

                        if isParent:
                            feature.setParentFeature(self.__prevFeature)

                        self.__prevFeature = feature
                    else:
                        obj.setPos(event.tilePos)
                        obj.show()
                    print "Placed " + self.__objectName + " @ " + `event.tilePos`
                    return True
            elif event.rmouse:
                feature = tile.getFeature(EventFeature.FEATURE_NAME)
                if feature:
                    if feature.getEventName() == SpearTrap.FEATURE_NAME:
                        leverTiles = self.world.getTilesWithEventFeatureParent(feature)
                        for leverTile in leverTiles:
                            leverTile.removeFeature(EventFeature.FEATURE_NAME)

                    feature.destroy()
                    tile.removeFeature(EventFeature.FEATURE_NAME)
        return False

    def __onObjectButtonBackgroundHoverStateChanged(self, btnBG, hovering):
        btnBG.fillcolor = ("999999" if hovering else "444444")

    def __onObjectButtonClicked(self, btn):
        self.setPlaceable(btn.objectName)
