####################################################################################
#
# Tilemap Editor for MazeRunner Game
#
####################################################################################

from EdConstants import *

import os, sys, inspect
sys.path.insert(1, os.path.join(sys.path[0], GAME_DIR))

from abc import ABCMeta, abstractmethod
from libavg import avg, app, player
from Game import Game
from Tile import *
from event.SpearTrap import *
from event.Portal import *
from Direction import *
from EdUtil import *
import Util
import math

##############################################################################################

class Placeable(object):
    def __init__(self, name, rawHref, world):
        self.__name = name
        self.__world = world
        self.__pos = (0,0)
        self.__href = rawHref

        href = self.getHref()
        self.__node = avg.ImageNode(href=href, active=False, sensitive=False, parent=world)
        self.__helper = avg.ImageNode(href=href, active=False, sensitive=False, opacity=0.6, parent=world)
        world.scaleNodeToTileSize(self.__node)
        world.scaleNodeToTileSize(self.__helper)

    def getName(self):
        return self.__name


    def getHref(self):
        return self.__href.replace('{tileset}', self.__world.tileset)

    def getRawHref(self):
        return self.__href

    def setRawHref(self, href):
        self.__href = href
        self.__node.href = self.getHref()

    # Recompiles raw href and updates the node and helper node's href
    def updateHref(self):
        self.setRawHref(self.getRawHref())


    def getWorld(self):
        return self.__world


    def getPos(self):
        return self.__pos

    # pos - Tuple giving the tile position
    def setPos(self, pos):
        self.__pos = pos
        tile = self.__world.getTile(pos)
        if tile:
            self.__node.pos = tile.pos
            self.__node.angle = math.radians(tile.rotation)
            self.show()
        else:
            self.hide()

    # tile-position
    pos = property(getPos, setPos)


    def show(self, show=True):
        self.__node.active = show
        bringNodeToTop(self.__node)

    def hide(self):
        self.show(False)

    def showHelper(self, show=True):
        self.__helper.active = show
        self.__helper.elementoutlinecolor = ""
        bringNodeToTop(self.__helper)

    def hideHelper(self):
        self.showHelper(False)

    def setHelperOutlineColor(self, color):
        self.__helper.elementoutlinecolor = color

    # pos - tilepos
    def setHelperPos(self, pos):
        tile = self.__world.getTile(pos)
        if tile:
            self.__helper.pos = tile.pos
            self.__helper.angle = math.radians(tile.rotation)
            self.__helper.elementoutlinecolor = ('00FF00' if tile.texture != 'ground' else 'FF0000')

    def bringToTop(self):
        bringNodeToTop(self.__node)
        bringNodeToTop(self.__helper)

    def serialize(self):
        return serializeVector(self.__pos)

    def deserialize(self, data):
        self.setPos(deserializeVector(data))


# Do not confuse with TileEvent - we want to have our own lightweight feature here
class EventFeature(TileFeature):
    FEATURE_NAME = "event_feature"

    def __init__(self, tile, eventName, prototypeObject):
        super(EventFeature, self).__init__(tile)
        self.__parentFeature = None
        self.__eventName = eventName
        self.__prototypeObject = prototypeObject
        self.__world = prototypeObject.getWorld()
        self.__helper = avg.ImageNode(
            pos=tile.pos,
            angle=math.radians(tile.rotation),
            size=tile.size,
            href=prototypeObject.getHref(),
            elementoutlinecolor="FF00FF",
            sensitive=False,
            parent=prototypeObject.getWorld())
        self.__parentAssocHelper = avg.LineNode(
            color="FF00FF",
            opacity=0.5,
            strokewidth=2,
            active=False,
            parent=prototypeObject.getWorld())

    def onDestroy(self):
        self.__prototypeObject = None
        if self.__helper:
            self.__helper.unlink(True)
            self.__helper = None
        if self.__parentAssocHelper:
            self.__parentAssocHelper.unlink(True)
            self.__parentAssocHelper = None

    def getFeatureName(self):
        return EventFeature.FEATURE_NAME

    def getEventName(self):
        return self.__eventName

    def getParentFeature(self):
        return self.__parentFeature

    # Sets the parent EventFeature
    def setParentFeature(self, parentFeature):
        self.__parentFeature = parentFeature
        if parentFeature:
            tile = self.getTile()
            parentTile = parentFeature.getTile()
            self.__parentAssocHelper.active = True
            self.__parentAssocHelper.pos1 = tile.pos + 0.5 * tile.size
            self.__parentAssocHelper.pos2 = parentTile.pos + 0.5 * parentTile.size

    def hasParent(self):
        return self.__parentFeature is not None

    def onPrototypeHrefChanged(self):
        if self.__prototypeObject:
            self.__helper.href = self.__prototypeObject.getHref()

    def onTileTransformed(self):
        self.__helper.pos = self.getTile().pos
        self.setParentFeature(self.__parentFeature)

class World(ScalableDivNode):
    # tileSz - avg.Point2D in pixels
    # worldSz - tuple in number of tiles per coordinate
    def __init__(self, parentNode, tileSz, worldSz, tileset, onTileClicked):
        super(World, self).__init__()
        self.registerInstance(self, parentNode)

        self.__origGroundTexture = Tile.textures['ground']
        Tile.textures['ground'] = '../../editor/empty.png'

        self.__tileset = tileset

        self.__tileSz = tileSz
        Tile.setTileSize(tileSz.x, tileSz.y)

        self.__worldSz = worldSz
        self.__worldSzPx = avg.Point2D(worldSz) * tileSz

        self.__deskSzPx = self.__worldSzPx + avg.Point2D(Game.deskBorderSz) * 2
        self.__deskTextureSz = avg.Point2D(DESK_TEXTURE_SIZE)
        self.__desk = avg.RectNode(
            pos=-self.__deskSzPx * 0.5,
            size=self.__deskSzPx,
            filltexhref=ASSET_DIR + "/frame.png",
            fillopacity=1.0,
            strokewidth=0,
            parent=self)

        self.__backgroundTextureSz = avg.Point2D(BACKGROUND_TEXTURE_SIZE)
        self.__background = avg.RectNode(
            pos=-self.__worldSzPx * 0.5,
            size=self.__worldSzPx,
            filltexhref=ASSET_DIR + "/" + self.__tileset + "/" + self.__origGroundTexture,
            fillopacity=1.0,
            strokewidth=0.0,
            parent=self)

        self.__grid = avg.RectNode(
            pos=self.__background.pos,
            size=self.__background.size,
            filltexhref="grid-tile.png",
            fillopacity=1.0,
            strokewidth=0.0,
            parent=self)

        self.__tiles = {}
        self.__objects = {}

        self.__onTileClicked = onTileClicked

    def clear(self):
        for tilePos in self.__tiles:
            tile = self.__tiles[tilePos]
            tile.texture = "ground"
            eventFeature = tile.getFeature(EventFeature.FEATURE_NAME)
            if eventFeature:
                tile.removeFeature(EventFeature.FEATURE_NAME)

    def showGrid(self, show=True):
        self.__grid.active = show
    def hideGrid(self):
        self.showGrid(False)
    def isGridShown(self):
        return self.__grid.active

    def showBackground(self, show=True):
        self.__background.active = show
    def hideBackground(self):
        self.showBackground(False)
    def isBackgroundShown(self):
        return self.__background.active

    #-------------------------------------------------------------------------------------------

    def addObject(self, placeable):
        name = placeable.getName()
        if not name in self.__objects:
            self.__objects[name] = placeable

    def getObject(self, name):
        if name in self.__objects:
            return self.__objects[name]
        else:
            return None

    def getTilesWithEventFeatureParent(self, parentFeature):
        tiles = []
        for tilePos in self.__tiles:
            tile = self.__tiles[tilePos]
            feature = tile.getFeature(EventFeature.FEATURE_NAME)
            if feature and feature.getParentFeature() == parentFeature:
                tiles.append(tile)
        return tiles

    #-------------------------------------------------------------------------------------------

    def getTileSize(self):
        self.__tileSz

    # Scales the node so that it fits into a tile while preserving its aspect ratio
    def scaleNodeToTileSize(self, node):
        if node.width < node.height:
            node.height *= self.__tileSz.x / node.width
            node.width = self.__tileSz.x
        else:
            node.width *= self.__tileSz.y / node.height
            node.height = self.__tileSz.y


    def getWorldSize(self):
        return self.__worldSz

    # To properly create new tiles, the world is first scaled to normal scale (1.0) and
    # then re-scaled.
    # newsz - tuple in number of tiles per coordinate
    # onResizing - function(newsz) called when world was resized but not re-scaled again
    def setWorldSize(self, newsz, onResizing=None):
        self.__worldSz = newsz
        self.__worldSzPx = avg.Point2D(newsz) * self.__tileSz
        self.__deskSzPx = self.__worldSzPx + avg.Point2D(Game.deskBorderSz) * 2

        worldScale = self.scale
        self.scale = 1.0

        self.__background.size = self.__worldSzPx
        self.__background.pos = -self.__worldSzPx * 0.5

        self.__grid.size = self.__background.size
        self.__grid.pos = self.__background.pos

        self.__desk.size = self.__deskSzPx
        self.__desk.pos = -self.__deskSzPx * 0.5

        self.updateBackgroundTexcoords()

        # Clip existing tiles:
        remove = []
        for tile in self.__tiles:
            if tile[0] >= newsz[0] or tile[1] >= newsz[1]:
                remove.append(tile)

        for tilePos in remove:
            self.__tiles.pop(tilePos).unlink(True)

        # Add new tiles:
        for y in xrange(newsz[1]):
            for x in xrange(newsz[0]):
                if (x,y) in self.__tiles:
                    tile = self.__tiles[(x,y)]
                else:
                    tile = self.__tiles[(x,y)] = Tile('ground', self.__tileset, (0,0), self, False, None)
                    tile.subscribe(avg.Node.CURSOR_DOWN, lambda e, tile=tile: self.__onTileClicked(e, tile))

                tile.pos = avg.Point2D(x,y) * self.__tileSz - self.__worldSzPx * 0.5

                eventFeature = tile.getFeature(EventFeature.FEATURE_NAME)

        # We have to do another iteration to allow updating parent assoc helpers
        for pos in self.__tiles:
            eventFeature = self.__tiles[pos].getFeature(EventFeature.FEATURE_NAME)
            if eventFeature is not None:
                    eventFeature.onTileTransformed()

        if onResizing:
            onResizing(newsz)

        self.scale = worldScale

    def getWorldPixelSize(self):
        return self.__worldSzPx

    def updateBackgroundTexcoords(self):
        self.__background.filltexcoord1 = (0,0)
        print "backgroundTextureSz = " + `self.__backgroundTextureSz`
        self.__background.filltexcoord2 = self.__worldSzPx / self.__backgroundTextureSz

        self.__grid.filltexcoord1 = (0,0)
        self.__grid.filltexcoord2 = self.__worldSzPx / self.__tileSz

        self.__desk.filltexcoord1 = (0,0)
        self.__desk.filltexcoord2 = self.__deskSzPx / self.__deskTextureSz

    def getBackgroundTextureSize(self):
        return (int(self.__backgroundTextureSz.x),int(self.__backgroundTextureSz.y))
    def setBackgroundTextureSize(self, sz):
        self.__backgroundTextureSz = avg.Point2D(sz)
        self.updateBackgroundTexcoords()
    backgroundTextureSz = property(getBackgroundTextureSize, setBackgroundTextureSize)

    def getDeskTextureSize(self):
        return (int(self.__deskTextureSz.x),int(self.__deskTextureSz.y))
    def setDeskTextureSize(self, sz):
        self.__deskTextureSz = avg.Point2D(sz)
        self.updateBackgroundTexcoords()
    deskTextureSz = property(getDeskTextureSize, setDeskTextureSize)


    #-------------------------------------------------------------------------------------------

    # Returns the tile at the given tile position or None if not existing
    # pos - tuple giving tile position
    def getTile(self, (x,y)):
        if x >= 0 and y >= 0 and x < self.__worldSz[0] and y < self.__worldSz[1]:
            return self.__tiles[(x,y)]
        else:
            return None

    # Returns true if there is a tile at the given tile position (tuple)
    def tileExists(self, pos):
        return (pos in self.__tiles)

    # Returns Tile instance of the tile at given pixel pos or None if out of bounds
    def getTileAtPixelPos(self, p):
        if (p.x < 0 or p.x >= self.__worldSzPx.x) or (p.y < 0 or p.y >= self.__worldSzPx):
            return None
        tilePos = avg.Point2D(p.x // self.__tileSz.x, p.y // self.__tileSz.y)
        tilePos = (int(tilePos.x),int(tilePos.y))
        return self.__tiles[tilePos] if tilePos in self.__tiles else None

    # Returns the tile-position of the tile as a tuple or None if tile is None
    def getTilePos(self, tile):
        if tile is None:
            return None
        wpos = (tile.pos / self.scale) + self.__worldSzPx * 0.5
        return (int(round(wpos.x / self.__tileSz.x)), int(round(wpos.y / self.__tileSz.y)))

    def getFreeDirs(self, (x,y)):
        dirs = [Direction.UP, Direction.RIGHT, Direction.DOWN, Direction.LEFT]
        freeDirs = []
        for dir in dirs:
            tile = self.getTile((x + dir.tuple()[0], y + dir.tuple()[1]))
            if tile and tile.texture != 'ground':
                freeDirs.append(dir)
        return freeDirs

    # Returns a list of 4 tile-positions around given tile position: Up, Right, Down, Left
    @staticmethod
    def getNeighborTilePositions(tilePos):
        p = avg.Point2D(tilePos)
        return [p + (0,-1), p + (1,0), p + (0,1), p + (-1,0)]

    def getTileset(self):
        return self.__tileset

    def setTileset(self, tileset):
        self.__tileset = tileset
        self.__background.filltexhref = ASSET_DIR + "/" + self.__tileset + "/" + self.__origGroundTexture
        for objectName in self.__objects:
            self.__objects[objectName].updateHref()

        for tilePos in self.__tiles:
            tile = self.__tiles[tilePos]
            tile.tileset = tileset

            eventFeature = tile.getFeature(EventFeature.FEATURE_NAME)
            if eventFeature is not None:
                eventFeature.onPrototypeHrefChanged()

    tileset = property(getTileset, setTileset)

    #-------------------------------------------------------------------------------------------

    # Recalculates parameters of path tiles in the area around the given tile (tile-position)
    # hs - Half-size of the quadratic region (one side * 0.5). If <= 0, this function will update the whole world.
    def invalidateSurroundings(self, (tx,ty), hs=5):
        changes = []
        minReg = ((tx - hs, ty - hs) if hs > 0 else (0,0))
        maxReg = ((tx + hs + 1, ty + hs + 1) if hs > 0 else (self.__worldSz[0],self.__worldSz[1]))
        for x in range(minReg[0], maxReg[0]):
            for y in range(minReg[1], maxReg[1]):
                tile = self.getTile((x,y))
                if tile is None or tile.texture == 'ground':
                    continue

                freeDirs = self.getFreeDirs((x,y))
                numFreeDirs = len(freeDirs)
                tileType = 'ground'
                tileDirs = [] # default directions, defined by the asset
                if numFreeDirs == 1:
                    tileType = 'path_straight'
                    tileDirs = [Direction.DOWN]
                elif numFreeDirs == 2:
                    if abs(freeDirs[0].angle(freeDirs[1])) == 180:
                        tileType = 'path_straight'
                        tileDirs = [Direction.UP, Direction.DOWN]
                    else:
                        tileType = 'path_curve'
                        tileDirs = [Direction.RIGHT, Direction.DOWN]
                elif numFreeDirs == 3:
                    tileType = 'path_curve' # T-crossing
                    tileDirs = [Direction.LEFT, Direction.DOWN, Direction.RIGHT]
                elif numFreeDirs > 3:
                    tileType = 'path_curve' # 4-way crossing
                    tileDirs = [Direction.LEFT, Direction.UP, Direction.RIGHT, Direction.DOWN]

                freeDirs = [d.tuple() for d in freeDirs]
                tileDirs = [d.tuple() for d in tileDirs]
                angle = matchVectors(tileDirs, freeDirs)

                changes.append({
                    'pos': (x,y),
                    'angle': angle,
                    'texture': tileType
                })

        for change in changes:
            tile = self.getTile(change['pos'])
            tile.setRotation(change['angle'])
            if change['texture'] != 'ground':
                tile.texture = change['texture']

    #-------------------------------------------------------------------------------------------

    # Returns serialized world data as a dict
    # startPos - tuple giving tile position of where to start the path
    def serialize(self, startPos):
        path = []
        events = {}
        if startPos:
            # Traverse with an iterative DFS as long as there are untraversed path tiles
            open = [startPos]
            closed = []
            while (True):
                # Check for open orphan tiles
                if len(open) == 0:
                    for tilePos in self.__tiles:
                        if self.__tiles[tilePos].texture != 'ground' and not tilePos in closed:
                            open.append(tilePos)
                            break
                    if len(open) == 0:
                        break

                while len(open) > 0:
                    pos = open.pop()
                    closed.append(pos)
                    tileData = { 'x': pos[0],'y': pos[1] }
                    path.append(tileData)

                    tile = self.__tiles[pos]
                    eventFeature = tile.getFeature(EventFeature.FEATURE_NAME)
                    if eventFeature:
                        eventName = eventFeature.getEventName()
                        if eventFeature.hasParent():
                            continue
                        if not eventName in events:
                            events[eventName] = []

                        event = { 'x': pos[0], 'y': pos[1] }
                        if eventName == SpearTrap.FEATURE_NAME:
                            event['levers'] = []
                            leverTiles = self.getTilesWithEventFeatureParent(eventFeature)
                            for leverTile in leverTiles:
                                leverPos = self.getTilePos(leverTile)
                                event['levers'].append(serializeVector(leverPos))
                        elif eventName == Portal.FEATURE_NAME:
                            childTiles = self.getTilesWithEventFeatureParent(eventFeature)
                            if len(childTiles) > 0:
                                event['other'] = serializeVector(self.getTilePos(childTiles[0]))

                        events[eventName].append(event)

                    freeDirs = self.getFreeDirs(pos)
                    for freeDir in freeDirs:
                        freePos = (pos[0] + freeDir.tuple()[0], pos[1] + freeDir.tuple()[1])
                        if not freePos in closed:
                            open.append(freePos)

        return {
            'size': serializeVector(self.__worldSz),
            'backgroundTextureSize': serializeVector(self.__backgroundTextureSz),
            'path': path,
            'events': events
        }

    def deserialize(self, data):
        global BACKGROUND_TEXTURE_SIZE

        if not 'path' in data:
            print "Cannot deserialize world: Missing path data"
            return

        if 'backgroundTextureSize' in data:
            self.__backgroundTextureSz = avg.Point2D(deserializeVector(data['backgroundTextureSize']))

        if 'size' in data:
            worldSz = deserializeVector(data['size'])
        else:
            print "WARNING: Size missing in world data. Trying to determine..."
            worldSz = (1,1)
            for tileInfo in data['path']:
                if 'x' in tileInfo and 'y' in tileInfo:
                    if tileInfo['x'] >= worldSz[0]:
                        worldSz = (tileInfo['x'] + 1,worldSz[1])
                    if tileInfo['y'] >= worldSz[1]:
                        worldSz = (worldSz[0],tileInfo['y'] + 1)
            print "  -> calculated: " + `worldSz`

        # If the new world is bigger: clear first, then resize.
        # Otherwise, resize first, then clear
        newWorldBigger = (worldSz[0] * worldSz[1]) > (self.__worldSz[0] * self.__worldSz[1])
        if not newWorldBigger:
            self.setWorldSize(worldSz)

        self.clear()

        if newWorldBigger:
            self.setWorldSize(worldSz)

        # Load path tiles
        numInvalid = 0
        for tileInfo in data['path']:
            if 'x' in tileInfo and 'y' in tileInfo:
                self.__tiles[(tileInfo['x'],tileInfo['y'])].texture = 'path_straight'
            else:
                numInvalid += 1

        if numInvalid > 0:
            print "Warning: " + `numInvalid` + " corrupt tiles in world path data"

        self.invalidateSurroundings((0,0), 0) # invalidate whole world

        # Load events
        if 'events' in data:
            for eventName in data['events']:
                if eventName == SpearTrapLever.FEATURE_NAME:
                    continue

                for eventData in data['events'][eventName]:
                    pos = (eventData['x'],eventData['y'])
                    tile = self.getTile(pos)
                    feature = tile.getFeature(EventFeature.FEATURE_NAME)
                    if feature:
                        tile.removeFeature(EventFeature.FEATURE_NAME)
                    feature = EventFeature(tile, eventName, self.getObject(eventName))
                    tile.addFeature(feature)

                    if eventName == SpearTrap.FEATURE_NAME and 'levers' in eventData:
                        for leverPos in eventData['levers']:
                            leverTile = self.getTile(deserializeVector(leverPos))
                            leverFeature = leverTile.getFeature(EventFeature.FEATURE_NAME)
                            if leverFeature:
                                leverTile.removeFeature(leverFeature.getFeatureName())
                            leverFeature = EventFeature(leverTile, SpearTrapLever.FEATURE_NAME, self.getObject(SpearTrapLever.FEATURE_NAME))
                            leverFeature.setParentFeature(feature)
                            leverTile.addFeature(leverFeature)
                    elif eventName == Portal.FEATURE_NAME and 'other' in eventData:
                        otherTile = self.getTile(deserializeVector(eventData['other']))
                        if otherTile.hasFeature(EventFeature.FEATURE_NAME):
                            otherTile.removeFeature(EventFeature.FEATURE_NAME)

                        otherFeature = EventFeature(otherTile, Portal.FEATURE_NAME, self.getObject(Portal.FEATURE_NAME))
                        otherFeature.setParentFeature(feature)
                        otherTile.addFeature(otherFeature)


                    self.getTilesWithEventFeatureParent(feature)
        else:
            print "Note: No event data in world data"
